import argparse
import distutils.util
import gzip
import json
import pickle
import re

from philter_lite.philter import Philter


def main():
    # get input/output/filename
    help_str = """ Philter -- PHI filter for clinical notes """
    ap = argparse.ArgumentParser(description=help_str)
    ap.add_argument(
        "-i",
        "--input",
        default="./data/i2b2_notes/",
        help="Path to the directory or the file that contains the PHI note, the default is ./data/i2b2_notes/",
        type=str,
    )
    ap.add_argument(
        "-a",
        "--anno",
        default="./data/i2b2_anno/",
        help="Path to the directory or the file that contains the PHI annotation, the default is ./data/i2b2_anno/",
        type=str,
    )
    ap.add_argument(
        "-o",
        "--output",
        default="./data/i2b2_results/",
        help="Path to the directory to save the PHI-reduced notes in, the default is ./data/i2b2_results/",
        type=str,
    )
    ap.add_argument(
        "-f",
        "--filters",
        default="./configs/integration_1.json",
        help="Path to our config file, the default is ./configs/integration_1.json",
        type=str,
    )
    ap.add_argument(
        "-x",
        "--xml",
        default="./data/phi_notes.json",
        help="Path to the json file that contains all xml data",
        type=str,
    )
    ap.add_argument(
        "-c",
        "--coords",
        default="./data/coordinates.json",
        help="Path to the json file that contains the coordinate map data",
        type=str,
    )
    ap.add_argument(
        "--eval_output",
        default="./data/phi/",
        help="Path to the directory that the detailed eval files will be outputted to",
        type=str,
    )
    ap.add_argument(
        "-v",
        "--verbose",
        default=True,
        help="When verbose is true, will emit messages about script progress",
        type=lambda x: bool(distutils.util.strtobool(x)),
    )
    ap.add_argument(
        "-e",
        "--run_eval",
        default=True,
        help="When run_eval is true, will run our eval script and emit summarized results to terminal",
        type=lambda x: bool(distutils.util.strtobool(x)),
    )
    ap.add_argument(
        "-t",
        "--freq_table",
        default=False,
        help="When freqtable is true, will output a unigram/bigram frequency table of all note words and their PHI/non-PHI counts",
        type=lambda x: bool(distutils.util.strtobool(x)),
    )
    ap.add_argument(
        "-n",
        "--initials",
        default=True,
        help="When initials is true, will include initials PHI in recall/precision calculations",
        type=lambda x: bool(distutils.util.strtobool(x)),
    )
    ap.add_argument(
        "--stanfordner",
        default="/usr/local/stanford-ner/",
        help="Path to Stanford NER, the default is /usr/local/stanford-ner/",
        type=str,
    )
    ap.add_argument(
        "--outputformat",
        default="asterisk",
        help='Define format of annotation, allowed values are "asterisk", "i2b2". Default is "asterisk"',
        type=str,
    )
    ap.add_argument(
        "--ucsfformat",
        default=False,
        help="When ucsfformat is true, will adjust eval script for slightly different xml format",
        type=lambda x: bool(distutils.util.strtobool(x)),
    )
    ap.add_argument(
        "--prod",
        default=False,
        help="When prod is true, this will run the script with output in i2b2 xml format without running the eval script",
        type=lambda x: bool(distutils.util.strtobool(x)),
    )
    ap.add_argument(
        "--cachepos",
        default=None,
        help="Path to a directoy to store/load the pos data for all notes. If no path is specified then memory caching will be used.",
        type=str,
    )

    args = ap.parse_args()
    run_eval = args.run_eval
    verbose = args.verbose

    if args.prod:
        run_eval = False
        verbose = False
        outputformat = "i2b2"
        # filters = "./configs/philter_alpha.json"
        philter_config = {
            "verbose": verbose,
            "run_eval": run_eval,
            "finpath": args.input,
            "foutpath": args.output,
            "outformat": outputformat,
            "filters": args.filters,
            "cachepos": args.cachepos,
        }

    else:
        philter_config = {
            "verbose": args.verbose,
            "run_eval": args.run_eval,
            "freq_table": args.freq_table,
            "initials": args.initials,
            "finpath": args.input,
            "foutpath": args.output,
            "outformat": args.outputformat,
            "ucsfformat": args.ucsfformat,
            "anno_folder": args.anno,
            "filters": args.filters,
            "xml": args.xml,
            "coords": args.coords,
            "eval_out": args.eval_output,
            "cachepos": args.cachepos,
            "stanford_ner_tagger": {
                "classifier": args.stanfordner
                + "classifiers/english.all.3class.distsim.crf.ser.gz",
                "jar": args.stanfordner + "stanford-ner.jar",
                "download": True,
            },
        }

    if verbose:
        print("RUNNING ", philter_config["filters"])

    filterer = Philter(philter_config)

    # map any sets, pos and regex groups we have in our config
    filterer.map_coordinates()

    # transform the data
    # Priority order is maintained in the pattern list
    filterer.transform()

    # evaluate the effectiveness
    if run_eval and args.outputformat == "asterisk":
        filterer.eval(
            philter_config,
            in_path=args.output,
            anno_path=args.anno,
            anno_suffix=".txt",
            fn_output="data/phi/fn.txt",
            fp_output="data/phi/fp.txt",
            summary_output="./data/phi/summary.json",
            phi_matcher=re.compile("\*+"),
            pre_process=r":|\,|\-|\/|_|~",  # characters we're going to strip from our notes to analyze against anno
            only_digits=False,
            pre_process2=r"[^a-zA-Z0-9]",
            punctuation_matcher=re.compile(r"[^a-zA-Z0-9\*]"),
        )


# error analysis

if __name__ == "__main__":
    main()



import argparse
import sys
import xml.etree.ElementTree as ET

sys.path
sys.path.append("/usr/local/lib/python2.7/site-packages/")
import os

import xmltodict

# This script removes PHI tags that are not PHI (according to HIPAA) from i2b2 annotations


def extractXML(directory, filename):
    tree = ET.parse(directory + "/" + filename)
    root = tree.getroot()
    xmlstr = ET.tostring(root, encoding="utf8", method="xml")
    # print xmlstr + "\n \n"
    xml_dict = xmltodict.parse(xmlstr)["deIdi2b2"]
    text = xml_dict["TEXT"]
    tags_dict = xml_dict["TAGS"]
    return text, tags_dict, xmlstr


def delete_annotation(xml_file, phi_type, tag_to_delete):
    # print tag_to_delete
    remove_line_if = 'text="' + tag_to_delete + '"'

    for line in xml_file.split("\n"):
        # print remove_line_if
        if remove_line_if in line:
            remove_line = line + "\n"
            print(remove_line)
            xml_file = xml_file.replace(remove_line, "")

    return xml_file


def fix_dates(xml_file, text):
    # remove year
    PHI_type = "DATE"
    date = text
    if date.isdigit():
        if len(date) == 4 and (1000 <= date <= 3000):
            delete_annotation(xml_file, PHI_type, text)

    # remove season
    seasons = ["spring", "winter", "fall", "summer", "autumn"]
    if date.lower() in seasons:
        delete_annotation(xml_file, PHI_type, text)

    # remove day of week
    days = [
        "Monday",
        "Tuesday",
        "Wednesday",
        "Thursday",
        "Friday",
        "Saturday",
        "Sunday",
        "Mon",
        "Tues",
        "Wed",
        "Thurs",
        "Fri",
        "Sat",
        "Sun",
    ]
    if date in days:
        delete_annotation(xml_file, PHI_type, text)

    return xml_file


def remove_abbrevs(xml_file, text, phi_type):
    # question -- what do we do about the username field

    if len(text) < 4 and str(text).isupper():
        delete_annotation(xml_file, phi_type, text)
    return xml_file


def remove_ids(xml_file, text):
    # three sub-categories:
    ## IDNUM
    ## MEDICALRECORD
    ## DEVICE

    phi_type = "ID"
    if len(text) < 5:
        xml_file = delete_annotation(xml_file, PHI_type, text)

    return xml_file


def remove_countries(xml_file, text, phi_type):
    if phi_type == "COUNTRY":
        xml_file = delete_annotation(xml_file, phi_type, text)
    return xml_file


def remove_age_under_90(xml_file, text, phi_type, filename):
    standardized_age = text.split("y")[0]
    non_numeric_ages = [
        "18 month",
        "20's",
        "32y5.7m",
        "56y0.5m",
        "50's",
        "50s",
        "60's",
        "60's",
        "60's",
        "60's",
        "60s",
        "60s",
        "60s",
        "70's",
        "70's",
        "70s",
        "80's",
        "80's",
        "80's",
        "80's",
        "80's",
        "80's",
        "80's",
        "80's",
        "80's",
        "80's",
        "80s",
        "80s",
        "80s",
        "80s",
        "80s",
        "82nd",
        "Sevent",
        "sevent",
    ]
    if standardized_age in non_numeric_ages:
        xml_file = delete_annotation(xml_file, phi_type, text)
    elif isinstance(standardized_age, int) and int(standardized_age) < 90:
        xml_file = delete_annotation(xml_file, phi_type, text)
    elif standardized_age not in non_numeric_ages and int(standardized_age) < 90:
        xml_file = delete_annotation(xml_file, phi_type, text)
    # else:
    # 	print filename,text
    return xml_file

    ## Notes: Kept the label in the following notes of 90
    # 131-02.xml, 310-02.xml, 310-04.xml, 373-04.xml, 373-05.xml, 215-02.xml
    # 194-02.xml, 136-05.xml


def remove_hospitals(xml_file, text, phi_type):
    if phi_type == "HOSPITAL":
        xml_file = delete_annotation(xml_file, phi_type, text)
    return xml_file


def main():

    ap = argparse.ArgumentParser()
    ap.add_argument(
        "-i",
        "--input",
        required=True,
        help="Path to the folder with the original i2b2 xml files (testing-PHI-Gold-fixed)",
    )
    ap.add_argument(
        "-o", "--output", required=True, help="Path to save the curated xml files to"
    )
    args = ap.parse_args()

    # Loop through xml_files
    input_dir = args.input
    output_dir = args.output

    # Make the output directroy if it doesn't exist already
    try:
        os.makedirs(output_dir)
    except OSError:
        print("Output directory already exists.")

    for filename in os.listdir(input_dir):
        print("\nCurating: " + filename)

        # example for year
        # filename = "113-02.xml"

        # example for season
        # filename = "119-03.xml"

        # example for weekday
        # filename ="110-04.xml"

        # example for abbrev
        # filename = "236-01.xml"

        # re-indent
        text, tags_dict, xmlstr = extractXML(input_dir, filename)
        # print str(tags_dict) + '\n \n'

        for key, value in tags_dict.iteritems():
            # Note:  Value can be a list of like phi elements
            # 		or a dictionary of the metadata about a phi element

            if isinstance(value, list):
                for final_value in value:
                    # do checks
                    text = final_value["@text"]
                    phi_type = final_value["@TYPE"]

                    if phi_type == "DATE":
                        xmlstr = fix_dates(xmlstr, text)
                        # elif phi_type == "NAME" or phi_type == "DOCTOR":
                    # 	xmlstr = remove_abbrevs(xmlstr,text,phi_type)
                    # elif phi_type == "ID":
                    # 	xmlstr = remove_ids(xmlstr,text)
                    elif phi_type == "COUNTRY":
                        xmlstr = remove_countries(xmlstr, text, phi_type)
                    elif phi_type == "AGE":
                        xmlstr = remove_age_under_90(xmlstr, text, phi_type, filename)
                    # elif phi_type == "HOSPITAL":
                    # 	xmlstr = remove_hospitals(xmlstr,text,phi_type)
            else:
                final_value = value
                text = final_value["@text"]
                phi_type = final_value["@TYPE"]

                if phi_type == "DATE":
                    xmlstr = fix_dates(xmlstr, text)
                # elif phi_type == "NAME" or phi_type == "DOCTOR":
                # 	xmlstr = remove_abbrevs(xmlstr,text,phi_type)
                # elif phi_type == "ID":
                # 		xmlstr = remove_ids(xmlstr,text)
                elif phi_type == "COUNTRY":
                    xmlstr = remove_countries(xmlstr, text, phi_type)
                elif phi_type == "AGE":
                    xmlstr = remove_age_under_90(xmlstr, text, phi_type, filename)
                # elif phi_type == "HOSPITAL":
                # 	xmlstr = remove_hospitals(xmlstr,text,phi_type)

        # here we write back out the updated XML File to a new directory
        output_file = output_dir + filename
        with open(output_file, "w") as text_file:
            text_file.write(xmlstr)


if __name__ == "__main__":
    main()



# Adds variables in to regex to create complex regex

import json
import os

# Define regex variables

day_name = "(S|s)un(day)?(s)?|SUN(DAY)?(S)?|(M|m)on(day)?(s)?|MON(DAY)?(S)?|(T|t)ues(day)?(s)?|Tue|TUES(DAY)?(S)?|(W|w)ed(nesday)?(s)?|WED(NESDAY)?(S)?|(T|t)hurs(day)?(s)?|Thu|THURS(DAY)?(S)?|(F|f)ri(day)?(s)?|FRI(DAY)?(S)?|(S|s)at(urday)?(s)?|SAT(URDAY)?(S)?"

month_name = "(J|j)an(uary)?|JAN(UARY)?|(F|f)eb(ruary)?|FEB(RUARY)?|(M|m)ar(ch)?|MAR(CH)?|(A|a)pr(il)?|APR(IL)?|May|MAY|(J|j)un(e)?|JUN(E)?|(J|j)ul(y)?|JUL(Y)?|(A|a)ug(ust)?|AUG(UST)?|(S|s)ep(tember)?|SEP(TEMBER)?|SEPT|Sept|(O|o)ct(ober)?|OCT(OBER)?|(N|n)ov(ember)?|NOV(EMBER)?|(D|d)ec(ember)?|DEC(EMBER)?"

day_numbering = "1st|2nd|3rd|4th|5th|6th|7th|8th|9th|10th|11th|12th|13th|14th|15th|16th|17th|18th|19th|20th|21st|22nd|23rd|24th|25th|26th|27th|28th|29th|30th|31st|32nd|33rd|34th|35th|36th|37th|38th|39th|40th|41st|42nd|43rd|44th|45th|46th|47th|48th|49th|50th"

seasons = "(S|s)pring|SPRING|(F|f)all|FALL|(A|a)utumn|AUTUMN|(W|w)inter|WINTER|(S|s)ummer|SUMMER|(C|c)hristmas|(N|n)ew (Y|y)ear's (E|e)ve"

# old_address_indicator = "alley|alley|ally|aly|anex|annex|annx|anx|apartment|apt|arcade|arcade|avenue|ave|aven|avenu|avenue|avn|avnue|bayou|bayou|beach|beach|bend|bnd|bluff|bluf|bluff|bluffs|bottom|btm|bottm|bottom|boulevard|boul|boulevard|blvd|boulv|box|branch|brnch|branch|bridge|brg|bridge|brook|brook|brooks|burg|burgs|bypass|bypa|bypas|bypass|byps|camp|cp|cmp|canyon|canyon|cnyn|cape|cpe|causeway|causwa|cswy|center|cent|center|centr|centre|cnter|cntr|ctr|centers|circle|circ|circl|circle|crcl|crcle|circles|cliff|cliff|cliffs|cliffs|club|club|common|commons|corner|corner|corners|cors|crse|court|ct|courts|cts|cove|cv|coves|creek|crk|crescent|cres|crsent|crsnt|crest|crossing|crssng|xing|crossroad|crossroads|curve|dale|dl|dam|dm|divide|divide|dv|dvd|drive|driv|dr|drv|drives|east|estate|estate|estates|ests|expressway|expr|express|expressway|expw|expy|extension|extension|extn|extnsn|extensions|fall|falls|fls|ferry|frry|fry|field|fld|fields|flds|flat|flt|flats|flts|ford|frd|fords|forest|forests|frst|forge|forge|frg|forges|fork|frk|forks|frks|fort|frt|ft|freeway|freewy|frway|frwy|fwy|garden|gardn|grden|grdn|gardens|gdns|grdns|gateway|gatewy|gatway|gtway|gtwy|glen|gln|glens|green|grn|greens|grove|grove|grv|groves|harbor|harbor|harbr|hbr|hrbor|harbors|haven|hvn|heights|hts|highway|highwy|hiway|hiwy|hway|hwy|hill|hl|hills|hls|hollow|hollow|hollows|holw|holws|inlet|island|island|islnd|islands|islnds|iss|isle|isles|junction|jction|jctn|junction|junctn|juncton|junctions|jcts|junctions|key|ky|keys|kys|knoll|knol|knoll|knolls|knolls|lake|lake|lakes|lakes|land|landing|lndg|lndng|lane|ln|light|light|lights|loaf|loaf|lock|lock|locks|locks|lodge|ldge|lodg|lodge|loop|loops|mall|manor|manor|manors|mnrs|meadow|meadows|mdws|meadows|medows|mews|mill|mills|mission|mssn|motorway|mount|mt|mount|mountain|mntn|mountain|mountin|mtin|mtn|mountains|mountains|neck|neck|north|orchard|orchard|orchrd|oval|ovl|overpass|park|prk|parks|parkway|parkwy|pkway|pkwy|pky|parkways|pkwys|pass|passage|path|paths|pike|pikes|pine|pines|pnes|place|plain|pln|plains|plns|plaza|plz|plza|point|pt|points|pts|port|prt|ports|prts|prairie|prairie|prr|radial|radial|radiel|radl|ramp|ranch|ranches|rnch|rnchs|rapid|rpd|rapids|rpds|rest|rst|ridge|rdge|ridge|ridges|ridges|river|river|rvr|rivr|road|road|rd|roads|rds|route|row|rue|run|shoal|shoal|shoals|shoals|shore|shore|shr|shores|shores|shrs|skyway|south|spring|spng|spring|sprng|springs|spngs|springs|sprngs|spur|spurs|square|sqr|sqre|squ|square|squares|squares|station|station|statn|stn|stravenue|strav|straven|stravenue|stravn|strvn|strvnue|stream|streme|strm|street|strt|st|str|streets|suite|ste|summit|sumit|sumitt|summit|terrace|terr|terrace|throughway|trace|traces|trce|track|tracks|trak|trk|trks|trafficway|trail|trails|trl|trls|trailer|trlr|trlrs|tunnel|tunl|tunls|tunnel|tunnels|tunnl|turnpike|turnpike|turnpk|underpass|union|union|unions|valley|vally|vlly|vly|valleys|vlys|viaduct|via|viadct|viaduct|view|vw|views|vws|village|villag|village|villg|villiage|vlg|villages|vlgs|ville|vl|vista|vist|vista|vst|vsta|walk|walks|wall|way|way|ways|well|wells|west|wls"

# Modified address_indicator on 08/24/18
address_indicator = "alley|alley|ally|aly|anex|annex|annx|anx|apartment|apt|avenue|ave|aven|avenu|avenue|avn|avnue|bayou|bayou|beach|beach|bnd|bluff|bluf|bluff|bluffs|boulevard|boul|boulevard|blvd|boulv|box|branch|brnch|branch|bridge|brg|bridge|brook|brook|brooks|burg|burgs|byps|canyon|canyon|cnyn|cape|cpe|causeway|causwa|cswy|cent|centr|centre|cnter|cntr|ctr|circle|circ|circl|circle|crcl|crcle|circles|cliff|cliff|cliffs|cliffs|club|club|commons|corner|corner|corners|cors|crse|court|ct|courts|cts|cove|cv|coves|creek|crk|crescent|cres|crsent|crsnt|crest|crossing|crssng|xing|crossroad|crossroads|dale|dl|dam|divide|divide|dv|dvd|drive|driv|dr|drv|drives|east|estate|estate|estates|ests|expressway|expr|express|expressway|expw|expy|extn|extnsn|fls|ferry|frry|fry|flat|flt|flats|flts|ford|frd|fords|forest|forests|frst|forge|forge|frg|forges|fork|frk|forks|frks|fort|frt|ft|freeway|freewy|frway|frwy|fwy|garden|gardn|grden|grdn|gardens|gdns|grdns|gateway|gatewy|gatway|gtway|gtwy|glen|gln|glens|green|grn|greens|grove|grove|grv|groves|harbor|harbor|harbr|hbr|hrbor|harbors|haven|hvn|heights|hts|highway|highwy|hiway|hiwy|hway|hwy|hill|hl|hills|hls|hollow|hollow|hollows|holw|holws|iss|isle|isles|knoll|knol|knoll|knolls|knolls|lake|lake|lakes|lakes|land|landing|lndg|lndng|lane|ln|loaf|loaf|lock|lock|locks|locks|ldge|lodg|lodge|loop|loops|mall|manor|manor|manors|mnrs|meadow|meadows|mdws|meadows|medows|mews|mill|mills|mission|mssn|motorway|north|orchard|orchard|orchrd|oval|ovl|overpass|park|prk|parks|parkway|parkwy|pkway|pkwy|pky|parkways|pkwys|pass|passage|pike|pikes|pine|pines|pnes|place|plain|pln|plains|plns|plaza|plz|plza|port|prt|ports|prts|prairie|prairie|prr|ramp|ranch|ranches|rnch|rnchs|rapids|rpds|rst|ridge|rdge|ridge|ridges|ridges|river|river|rvr|rivr|road|road|rd|roads|rds|route|row|run|shoal|shoal|shoals|shoals|shore|shore|shr|shores|shores|shrs|skyway|south|spring|spng|spring|sprng|springs|spngs|springs|sprngs|spur|spurs|square|sqr|sqre|squ|square|squares|squares|station|station|statn|stn|stravenue|strav|straven|stravenue|stravn|strvn|strvnue|stream|streme|strm|street|strt|st|str|streets|suite|summit|sumit|sumitt|summit|terrace|terr|terrace|throughway|trce|track|tracks|trak|trk|trks|trafficway|trail|trails|trl|trls|trailer|trlr|trlrs|tunnel|tunl|tunls|tunnel|tunnels|tunnl|turnpike|turnpike|turnpk|underpass|union|union|unions|valley|vally|vlly|vly|valleys|vlys|viaduct|viadct|viaduct|village|villag|village|villg|villiage|vlg|villages|vlgs|ville|vl|vista|vist|vista|vst|vsta|way|way|ways|west|wls"

state_names = "(A|a)rizona|AZ|(V|v)irginia|VA|(M|m)innesota|MN|(A|a)laska|AK|(N|n)ew (Y|y)ork|NY|(T|t)exas|TX|(V|v)ermont|VT|(U|u)tah|UT|(N|n)ew (J|j)ersey|NJ|(N|n)orth (D|d)akota|ND|(S|s)outh (D|d)akota|SD|(M|m)issouri|MO|(W|w)ashington (D|d).(C|c).|(G|g)eorgia|GA|(M|m)assachusetts|MA|(P|p)uerto (R|r)ico|(M|m)ichigan|MI|(I|i)owa|IA|(N|n)orth (C|c)arolina|NC|(S|s)outh (C|c)arolina|SC|(N|n)evada|NV|(C|c)olorado|CO|(O|o)hio|OH|(H|h)awaii|HI|(N|n)ebraska|NE|(N|n)ew (H|h)ampshire|NH|(W|w)ashington|WA|(T|t)ennessee|TN|(A|a)rkansas|AR|(L|l)ouisiana|LA|(M|m)ississippi|MS|(O|o)regon|OR|(A|a)labama|AL|(W|w)yoming|WY|(W|w)isconsin|WI|(O|o)klahoma|OK|(F|f)lorida|FL|(R|r)hode (I|i)sland|RI|(I|i)ndiana|IN|(C|c)alifornia|CA|(K|k)ansas|KS|(D|d)elaware|DE|(M|m)aryland|(I|i)daho|ID|(P|p)ennsylvania|PA|(K|k)entucky|KY|(C|c)onnecticut|CT|(M|m)ontana|MT|(I|i)llinois|IL|(M|m)aine|ME"

full_numbering = "First|Second|Third|Fourth|Fifth|Sixth|Seventh|Eighth|Ninth|Tenth|Eleventh|Twelfth|Thirteenth|Fourteenth|Fifteenth|Sixteenth|Seventeenth|Eighteenth|Nineteenth|Twentieth"

# Get blacklisted names
# names_blacklist = json.loads(open("../../filters/blacklists/names_blacklist_ssfirst.json").read())
# person_names = ''
# for key in names_blacklist:
# 	person_names += key + '|'
# # Get rid of last pipe
# person_names = person_names[:-1]

# Do folder walk and transform each file
rootdir = "."
for subdir, dirs, files in os.walk(rootdir):
    for file in files:
        if ".txt" in file and "_transformed.txt" not in file and "catchall" not in file:
            filepath = os.path.join(subdir, file)
            # Get currnet file name and create transformed name
            file_root = file.split(".")[0]
            new_file_name = file_root + "_transformed.txt"
            new_filepath = os.path.join(subdir, new_file_name)
            # Open file
            regex = open(filepath, "r").read().strip()
            # Replace variables
            regex = (
                regex.replace('"""+month_name+r"""', month_name)
                .replace('"""+day_numbering+r"""', day_numbering)
                .replace('"""+day_name+r"""', day_name)
                .replace('"""+seasons+r"""', seasons)
                .replace('"""+address_indicator+r"""', address_indicator)
                .replace('"""+state_name+r"""', state_names)
                .replace('"""+full_numbering+r"""', full_numbering)
            )
            # Write new file
            with open(new_filepath, "w") as fin:
                fin.write(regex)



import argparse
import copy
import glob
import json
import os
import pickle
import re
from difflib import ndiff
from string import punctuation

from nltk import sent_tokenize, word_tokenize

"""
To run: python3 eval_scrubber_ucsf.py -p ./scrubber_out_ucsf_2k -a ./ucsf_anno -o ./ucsf_results

Compares the outputs of Scrubber de-identified notes to annotated notes to evaluate Scrubber performance. 
Provides
    - True and False Positive words found, and the total count of each
    - False Negative words found, and the total count
    - Precision and Recall scores

annotation.py returns a list of lists containing all words in the clinical note and their annotated phi-category: 
    - [[word1, phi-category],[word2, phi-category]]

phi-reducer.py returns a txt file (phi-reduced.txt) in which words that are phi have 'hopefully' been replaced with the safe word: **PHI**

evaluate.py
1. extracts all words from the annotation.py list for which the phi-category is 0 (not-phi) and adds them to a list (annot_list)
    - annot_list contains the True Negatives
2. extracts all non-**PHI** words from phi-reduced.txt and adds them to a list (phi_r_list)
3. get a count of all the **PHI** words that occurred in phi-reduced.txt (filtered_count)
4. Use ndiff() to compare annot_list to phi_r_list. Returns lines of strings containing the elements that were present in 1 list 
    but not in the other, with a symbol to identify which list element was present in. 
    - words that are in annot_list but not in phi_r_list are False Negatives (a phi-word got through)
    - words that are in phi_r_list but not in annot_list are False Positives (a non-phi word was filtered)
4. Filtered_Count = TP + FP - FN
    TP = Filtered_Count - FP + FN
    Use TP to calculate Precision and Recall

Returns:    (summary_dict.json) pickled file which is a dictionary of all FP and FN in all files that were processed. 
                Key: filename
                Values: list of FP words, list of FN words, Count of TP words
            (summary_text.txt) report containing the same information in summary_dict.json for each note and
                the precision/recall for each note and the counts of TP, FN, FP for all notes and the overall precision/recall for all notes
            (fn_tags_context.txt) text file that contains the word, context and most likely PHI type of all FNs in the notes
            (fp_tags_context.txt) text file that contains the word and context of all FNs in the notes

"""


def comparison(filename, file1path, file2path, allpositive_dict):

    ucsf_include_tags = [
        "Date",
        "Provider_Name",
        "Phone_Fax",
        "Patient_Name_or_Family_Member_Name",
        "Patient_Address",
        "Provider_Address_or_Location",
        "Provider_Certificate_or_License",
        "Patient_Medical_Record_Id",
        "Patient_Account_Number",
        "Patient_Social_Security_Number",
        "Patient_Vehicle_or_Device_Id",
        "Patient_Unique_Id",
        "Email",
        "URL_IP",
        "Patient_Biometric_Id_or_Face_Photo",
        "Patient_Certificate_or_License",
        "Age",
        "Patient_Initials",
        "Provider_Initials",
    ]

    summary_dict = {}
    file_context_dict = {"false_positives": [], "false_negatives": []}
    output = ""

    with open(file1path, "r") as fin:
        phi_reduced_note = fin.read()

    with open(file2path, "r") as fin:
        annotation_note = fin.read()
        # annotation_note = pickle.load(fin)
    # annotation_note = re.sub(r'[\/\-\:\~\_]', ' ', annotation_note)

    # get a list of sentences within the note , returns a list of lists  [[sent1],[sent2]]
    phi_reduced_note = re.sub(r"\[[A-Z]+\]", "", phi_reduced_note)
    phi_reduced_note = re.sub(r"\n", " ", phi_reduced_note)
    phi_reduced_note = re.sub(r"\#{5}\sDOCUMENT.*", "", phi_reduced_note)
    phi_reduced_note = re.sub(r"[\/\-\:\~\_]", " ", phi_reduced_note)
    phi_reduced_sentences = sent_tokenize(phi_reduced_note)

    # get a list of words within each sentence, returns a list of lists [[sent1_word1, sent1_word2, etc],[sent2_word1, sent2_word2, etc] ]
    phi_reduced_words = [word_tokenize(sent) for sent in phi_reduced_sentences]
    # a list of all words from the phi_reduced note: [word1, word2, etc]
    phi_reduced_list = [
        word for sent in phi_reduced_words for word in sent if word not in punctuation
    ]

    annotation_note = re.sub(r"[\/\-\:\~\_\n]", " ", annotation_note)
    annotation_sentences = sent_tokenize(annotation_note)
    annotation_words = [word_tokenize(sent) for sent in annotation_sentences]
    annotation_list = [
        word for sent in annotation_words for word in sent if word not in punctuation
    ]

    # Begin Step 1
    annot_list = [word for word in annotation_list if "*" not in word]
    for i in range(len(annot_list)):
        if annot_list[i][-1] in punctuation:
            annot_list[i] = annot_list[i][:-1]

    # annot_list = [word[0] for word in annotation_note if (word[1] == '0' or word[1] == '2')and word[0] != '']
    # for i in range(len(annot_list)):
    # if annot_list[i][-1] in punctuation:
    # annot_list[i] = annot_list[i][:-1]
    # check_set = {'of', 'any', 'for', 'spring', 'monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday', 'sunday'}
    check_set = {}

    # Begin Step 2
    phi_r_list = [word for word in phi_reduced_list if "*" not in word]
    for i in range(len(phi_r_list)):
        if phi_r_list[i][-1] in punctuation:
            phi_r_list[i] = phi_r_list[i][:-1]
    # Begin Step 3
    # filtered_count = [word for word in phi_reduced_list if '**PHI' in word]

    # filtered_count = len(filtered_count)
    summary_dict["false_positive"] = []
    summary_dict["false_negative"] = []
    # print(filtered_count)
    # print(annot_list)

    # marker_and_word are a string, eg "+ word" or "- word"
    # + means that the word appears in the first list but not in the second list
    # - means that the word appears in the second list but not in the first list
    # marker_and_word[2] is the first character of the word.
    # print(annot_list)
    minus_counter = 0
    plus_counter = 0
    # print(len(phi_r_list))
    # print(len(annot_list))
    # print('\n')

    for word_index, marker_and_word in enumerate(ndiff(phi_r_list, annot_list)):
        phi_word_index = word_index - plus_counter
        annot_word_index = word_index - minus_counter
        # print(word_index, marker_and_word)

        # Get context from PHI-reduced list
        if phi_word_index > 5 and phi_word_index < len(phi_r_list):
            phi_context = " ".join(phi_r_list[phi_word_index - 5 : phi_word_index + 6])
        elif phi_word_index > 5 and phi_word_index >= len(phi_r_list):
            phi_context = " ".join(phi_r_list[phi_word_index - 5 :])
        else:
            phi_context = " ".join(phi_r_list[: phi_word_index + 6])

        # Get context from annotated list
        if annot_word_index > 5 and annot_word_index < len(annot_list):
            annot_context = " ".join(
                annot_list[annot_word_index - 5 : annot_word_index + 6]
            )
        elif annot_word_index > 5 and annot_word_index >= len(annot_list):
            annot_context = " ".join(annot_list[annot_word_index - 5 :])
        else:
            annot_context = " ".join(annot_list[: annot_word_index + 6])

        # In the annotation list but not in the phi reduced list
        if marker_and_word[0] == "+" and re.findall(r"\w+", marker_and_word[2:]) != []:
            summary_dict["false_positive"].append([marker_and_word[2:], annot_context])
            file_context_dict["false_positives"].append(
                [marker_and_word[2:], annot_context]
            )
            plus_counter += 1
            # print(annot_word_index, marker_and_word, annot_context)
        # In the phi reduced list but not in the annotation list
        elif (
            marker_and_word[0] == "-" and re.findall(r"\w+", marker_and_word[2:]) != []
        ):
            summary_dict["false_negative"].append([marker_and_word[2:], phi_context])
            minus_counter += 1
            # print(phi_word_index, marker_and_word, phi_context)

    # print(summary_dict['false_positive'])
    temp_list = summary_dict["false_negative"]

    # We have our FN list. Now we need to sort it into different categories
    fn_dict = allpositive_dict[filename + ".txt.xml"]
    # Make a copy, so we can remove items from the dict without modifying the original
    fn_dict_copy = copy.deepcopy(fn_dict)

    # Make dictionary of tags to keep track of tagged FNs
    ucsf_category_fn_dict = {
        "Date": [],
        "Provider_Name": [],
        "Phone_Fax": [],
        "Age": [],
        "Patient_Name_or_Family_Member_Name": [],
        "Patient_Address": [],
        "Patient_Initials": [],
        "Provider_Address_or_Location": [],
        "Provider_Initials": [],
        "Provider_Certificate_or_License": [],
        "Patient_Medical_Record_Id": [],
        "Patient_Account_Number": [],
        "Patient_Social_Security_Number": [],
        "Patient_Vehicle_or_Device_Id": [],
        "Patient_Unique_Id": [],
        "Email": [],
        "URL_IP": [],
        "Patient_Certificate_or_License": [],
    }

    all_ucsf_fns = []

    ucsf_fps = 0

    # Determine 'true fns' by removing exlucing tags from the dict
    for tag in fn_dict:
        # First deal with ages
        if tag == "Age":
            for fn in fn_dict_copy[tag]:
                if fn.isdigit():
                    # If this age < 90, remove it from teh list of fns
                    if int(fn) < 90:
                        fn_dict_copy[tag].remove(fn)
                # Is the age a string?
                # Note that this won't catch all age FNs that are spelled ou
                # i.e., only the 'ninety' in 'ninety-five' will be marked as include
                # This won't affect our recall at all, but it will affect our precision a little
                # We will manually need to subtract theses from our FPs and add to our TPs
                else:
                    if "ninety" not in fn:
                        fn_dict_copy[tag].remove(fn)
        else:
            if tag not in ucsf_include_tags:
                fn_dict_copy.pop(tag)

    all_true_ucsf_fns = sum([len(fn_dict_copy[key]) for key in fn_dict_copy])

    # Remake dict copy
    fn_dict_copy = copy.deepcopy(fn_dict)

    # Iterate through false negatives
    for fn_list in temp_list:
        # Iterate through PHI categories and find a match (this will not be perfect)
        fn = fn_list[0]
        context = fn_list[1]
        for tag in fn_dict_copy:
            if fn in fn_dict_copy[tag]:
                # First make sure this is in our lsit of include tags
                if tag in ucsf_include_tags:
                    # If age, make sure the age is >90
                    if tag == "Age":
                        if fn.isdigit():
                            # If this age >= 90, add it to the list of fns
                            if int(fn) >= 90:
                                # Then add this word to the fn dict
                                ucsf_category_fn_dict[tag].append(fn)
                                all_ucsf_fns.append(fn)
                                file_context_dict["false_negatives"].append(
                                    [fn, tag, context]
                                )
                                # Remove this fn from the fn dict, to make sure we don't double count it
                                fn_dict_copy[tag].remove(fn)
                        else:
                            if "ninety" in fn:
                                # Then add this word to the fn dict
                                ucsf_category_fn_dict[tag].append(fn)
                                all_ucsf_fns.append(fn)
                                # Remove this fn from the fn dict, to make sure we don't double count it
                                fn_dict_copy[tag].remove(fn)
                    else:
                        # Then add this word to the fn dict
                        ucsf_category_fn_dict[tag].append(fn)
                        all_ucsf_fns.append(fn)
                        file_context_dict["false_negatives"].append([fn, tag, context])
                        # Remove this fn from the fn dict, to make sure we don't double count it
                        fn_dict_copy[tag].remove(fn)

    # print(file_context_dict)
    # Calculate true positives
    true_positive = all_true_ucsf_fns - len(all_ucsf_fns)
    summary_dict["true_positive"] = true_positive
    summary_dict["false_negative"] = all_ucsf_fns

    output = "Note: " + filename + "\n"
    # output += "Script filtered: " + str(filtered_count) + '\n'
    # print([item[0] for item in summary_dict['false_positive']])
    output += "True positive: " + str(true_positive) + "\n"
    output += (
        "False Positive: "
        + " ".join([item[0] for item in summary_dict["false_positive"]])
        + "\n"
    )
    output += "FP number: " + str(len(summary_dict["false_positive"])) + "\n"
    output += (
        "False Negative: "
        + " ".join([item[0] for item in summary_dict["false_negative"]])
        + "\n"
    )
    output += "FN number: " + str(len(summary_dict["false_negative"])) + "\n"
    if true_positive == 0 and len(summary_dict["false_negative"]) == 0:
        output += "Recall: N/A\n"
    else:
        output += (
            "Recall: {:.2%}".format(
                true_positive / (true_positive + len(summary_dict["false_negative"]))
            )
            + "\n"
        )
    # print(true_positive, len(summary_dict['false_positive']))
    if (true_positive == 0 and len(summary_dict["false_positive"]) == 0) or (
        true_positive < 0
    ):
        output += "Precision: N/A\n"
    else:
        output += (
            "Precision: {:.2%}".format(
                true_positive / (true_positive + len(summary_dict["false_positive"]))
            )
            + "\n"
        )

    output += "\n"
    # print(summary_dict)
    return summary_dict, output, ucsf_category_fn_dict, file_context_dict


def main():
    ap = argparse.ArgumentParser()
    ap.add_argument(
        "-p",
        "--phinote",
        required=True,
        help="Path to the phi reduced note, *.nphi.txt.",
    )
    ap.add_argument(
        "-a", "--annotation", required=True, help="Path to the annotated file, *.txt."
    )
    ap.add_argument(
        "-o",
        "--output",
        required=True,
        help="Path to save the summary pkl and statistics text.",
    )
    ap.add_argument(
        "-r",
        "--recursive",
        action="store_true",
        default=False,
        help="whether read files in the input folder recursively.",
    )
    args = ap.parse_args()

    file1path = args.phinote
    file2path = args.annotation
    foutpath = args.output
    if_recursive = args.recursive
    summary_dict_all = {}
    summary_text = ""
    phi_reduced_dict = {}
    annotation_dict = {}
    miss_file = []
    TP_all = 0
    FP_all = 0
    FN_all = 0
    processed_count = 0
    output = ""
    if_update = False

    allpositive_dict = json.loads(
        open("/data/muenzenk/nlm_scrubber/ucsf_detailed_anno_dict.json").read()
    )

    all_ucsf_fn_dict = {
        "Date": 0,
        "Provider_Name": 0,
        "Phone_Fax": 0,
        "Age": 0,
        "Patient_Name_or_Family_Member_Name": 0,
        "Patient_Address": 0,
        "Patient_Initials": 0,
        "Provider_Address_or_Location": 0,
        "Provider_Initials": 0,
        "Provider_Certificate_or_License": 0,
        "Patient_Medical_Record_Id": 0,
        "Patient_Account_Number": 0,
        "Patient_Social_Security_Number": 0,
        "Patient_Vehicle_or_Device_Id": 0,
        "Patient_Unique_Id": 0,
        "Email": 0,
        "URL_IP": 0,
        "Patient_Certificate_or_License": 0,
    }

    all_files_context_dict = {}

    # allpositive_dict = json.loads(open('/data/muenzenk/nlm_scrubber/anno_ucsf_dict.json').read())
    if os.path.isfile(file1path) != os.path.isfile(file2path):
        print("phi note input and annotation input should be both files or folders.")
    else:
        if os.path.isfile(file1path):
            head1, tail1 = os.path.split(file1path)
            head2, tail2 = os.path.split(file2path)
            file1name = ".".join(tail1.split(".")[:-1])
            file2name = ".".join(tail2.split(".")[:-1])
            if file1name != file2name:
                print("Please make sure the filenames are the same in both file.")
            else:
                summary_dict, output, file_fn_dict, file_context_dict = comparison(
                    file1name, file1path, file2path, allpositive_dict
                )
                all_files_context_dict[file1name] = file_context_dict
                summary_dict_all[file1name] = summary_dict
                summary_text += output
                if_update = True
                for tag in all_ucsf_fn_dict:
                    all_ucsf_fn_dict[tag] += len(file_fn_dict[tag])
        else:
            # reply = input('Please make sure all files are ready.'
            #             'Press Enter to process or others to quit.> ')
            # if reply == '':
            if if_recursive:
                for f in glob.glob(file1path + "/**/*.txt", recursive=True):
                    head, tail = os.path.split(f)
                    filename = ".".join(tail.split(".nphi.txt")[:-1])
                    # if filename != '':
                    # note_id = re.findall(r'\d+', tail)[0]
                    phi_reduced_dict[filename] = f
                    processed_count += 1
                for f in glob.glob(file2path + "/**/*.txt", recursive=True):
                    head, tail = os.path.split(f)
                    filename = ".".join(tail.split(".txt")[:-1])
                    # if re.findall(r'\d+', tail) != []:
                    #    note_id = re.findall(r'\d+', tail)[0]
                    annotation_dict[filename] = f
            else:
                for f in glob.glob(file1path + "/*.txt"):
                    head, tail = os.path.split(f)
                    filename = ".".join(tail.split(".nphi.txt")[:-1])
                    # if re.findall(r'\d+', tail) != []:
                    # note_id = re.findall(r'\d+', tail)[0]
                    phi_reduced_dict[filename] = f
                    processed_count += 1
                for f in glob.glob(file2path + "/*.txt"):
                    head, tail = os.path.split(f)
                    filename = ".".join(tail.split(".txt")[:-1])
                    # if re.findall(r'\d+', tail) != []:
                    #    note_id = re.findall(r'\d+', tail)[0]
                    annotation_dict[filename] = f
            # print(phi_reduced_dict)
            # print('\n')
            # print('\n')
            # print('\n')
            # print('\n')
            # print(annotation_dict)
            for i in phi_reduced_dict.keys():
                if i in annotation_dict.keys():
                    # print(phi_reduced_dict[i])
                    # print(annotation_dict[i])

                    summary_dict, output, file_fn_dict, file_context_dict = comparison(
                        i, phi_reduced_dict[i], annotation_dict[i], allpositive_dict
                    )
                    all_files_context_dict[i] = file_context_dict
                    summary_dict_all[i] = summary_dict
                    summary_text += output
                    if_update = True
                    for tag in all_ucsf_fn_dict:
                        all_ucsf_fn_dict[tag] += len(file_fn_dict[tag])
                else:
                    miss_file.append(phi_reduced_dict[i])

            print(
                "{:d} out of {:d} phi reduced notes have been compared.".format(
                    processed_count - len(miss_file), processed_count
                )
            )
            print(
                "{} files have not found corresponding annotation as below.".format(
                    len(miss_file)
                )
            )
            # print('\n'.join(miss_file)+'\n')
            if processed_count != 0:
                for k, v in summary_dict_all.items():
                    TP_all += v["true_positive"]
                    FP_all += len(v["false_positive"])
                    FN_all += len(v["false_negative"])

                output = "{} notes have been evaulated.\n".format(
                    processed_count - len(miss_file)
                )
                output += "True Positive in all notes: " + str(TP_all) + "\n"
                output += "False Positive in all notes: " + str(FP_all) + "\n"
                output += "False Negative in all notes: " + str(FN_all) + "\n"
                if TP_all == 0 and FN_all == 0:
                    output += "Recall: N/A\n"
                else:
                    output += "Recall: {:.2%}".format(TP_all / (TP_all + FN_all)) + "\n"
                if TP_all == 0 and FP_all == 0:
                    output += "Precision: N/A\n"
                else:
                    output += (
                        "Precision: {:.2%}".format(TP_all / (TP_all + FP_all)) + "\n"
                    )
                summary_text += output
        # else:
        #     print("Please re-run the script after all the files are ok.")

        print(output)
        print("\n")
        for tag in all_ucsf_fn_dict:
            print(tag + ": " + str(all_ucsf_fn_dict[tag]))
        if if_update:
            json.dump(
                summary_dict_all, open(foutpath + "/summary_dict.json", "w"), indent=4
            )
            with open(foutpath + "/summary_text.txt", "w") as fout:
                fout.write(summary_text)

        with open(foutpath + "/fn_tags_context.txt", "w") as fn_file:
            fn_file.write(
                "note_word"
                + "|"
                + "phi_tag"
                + "|"
                + "context"
                + "|"
                + "filename"
                + "\n"
            )
            # print(fn_tags_condensed_context)
            for file in all_files_context_dict:
                current_dict = all_files_context_dict[file]
                current_list_all = current_dict["false_negatives"]
                for current_list in current_list_all:
                    # print(current_list)
                    fn_file.write(
                        current_list[0]
                        + "|"
                        + current_list[1]
                        + "|"
                        + current_list[2]
                        + "|"
                        + file
                        + "\n"
                    )

        with open(foutpath + "/fp_tags_context.txt", "w") as fp_file:
            fp_file.write("note_word" + "|" + "context" + "|" + "filename" + "\n")
            for key in all_files_context_dict:
                current_dict = all_files_context_dict[key]
                current_list_all = current_dict["false_positives"]
                for current_list in current_list_all:
                    fp_file.write(
                        current_list[0] + "|" + current_list[1] + "|" + file + "\n"
                    )


if __name__ == "__main__":
    main()



import json
import os
import re

import nltk


def seq_eval(
    note_lst,
    anno_lst,
    text_matcher=re.compile(r"[a-zA-Z0-9]"),
    phi_matcher=re.compile(r"\*+"),
):
    """
        Compares two sequences item by item,
        returns generator which yields:
        classifcation, word

        classifications can be TP, FP, FN, TN
        corresponding to True Positive, False Positive, False Negative and True Negative
    """

    start_coordinate = 0
    for note_word, anno_word in list(zip(note_lst, anno_lst)):

        # Get coordinates
        note_word_stripped = re.sub(r"[^a-zA-Z0-9\*]+", "", note_word.strip())
        if len(note_word_stripped) == 0:
            # got a blank space or something without any characters or digits, move forward
            start_coordinate += len(note_word)
            continue

        if phi_matcher.search(anno_word):
            # this contains phi

            if note_word == anno_word:
                yield "TP", note_word, start_coordinate

            else:
                if text_matcher.search(anno_word):
                    # this is a complex edge case,
                    # the phi annotation has some characters *'ed, and some not,
                    # find the overlap and report any string of chars in anno as FP
                    # and any string of chars in note as FN
                    fn_words = []
                    fp_words = []

                    fn_chunk = []
                    fp_chunk = []
                    for n, a in list(zip(note_word, anno_word)):
                        if n == a:
                            # these characters match, clear our chunks
                            if len(fp_chunk) > 0:
                                fp_words.append("".join(fp_chunk))
                                fp_chunk = []
                            if len(fn_chunk) > 0:
                                fn_words.append("".join(fn_words))
                                fn_chunk = []

                            continue
                        if a == "*" and n != "*":
                            fn_chunk.append(n)
                        elif a != "*" and n == "*":
                            fp_chunk.append(a)

                    # clear any remaining chunks
                    if len(fp_chunk) > 0:
                        fp_words.append("".join(fp_chunk))
                    if len(fn_chunk) > 0:
                        fn_words.append("".join(fn_words))

                    # now drain the difference
                    for w in fn_words:
                        yield "FN", w, start_coordinate
                    for w in fp_words:
                        yield "FP", w, start_coordinate

                else:
                    # simpler case, anno word is completely blocked out except punctuation
                    yield "FN", note_word, start_coordinate

        else:
            if not note_word.isspace():
                # this isn't phi
                if note_word == anno_word:
                    yield "TN", note_word, start_coordinate
                else:
                    yield "FP", anno_word, start_coordinate

        # advance our start coordinate
        start_coordinate += len(note_word)


def eval(
    self,
    anno_path="data/i2b2_anno/",
    anno_suffix="_phi_reduced.ano",
    in_path="data/i2b2_results/",
    summary_output="data/phi/summary.json",
    fn_output="data/phi/fn.txt",
    fp_output="data/phi/fp.txt",
    punctuation_matcher=re.compile(r"[^a-zA-Z0-9\*]"),
):
    """ calculates the effectiveness of the philtering / extraction

        only_digits = <boolean> will constrain evaluation on philtering of only digit types
    """

    if not os.path.exists(anno_path):
        raise Exception("Anno Filepath does not exist", anno_path)
    if not os.path.exists(in_path):
        raise Exception("Input Filepath does not exist", in_path)

    summary = {
        "total_false_positives": 0,
        "total_false_negatives": 0,
        "total_true_positives": 0,
        "total_true_negatives": 0,
        "false_positives": [],  # non-phi words we think are phi
        "true_positives": [],  # phi words we correctly identify
        "false_negatives": [],  # phi words we think are non-phi
        "true_negatives": [],  # non-phi words we correctly identify
        "summary_by_file": {},
    }
    summary_coords = {"summary_by_file": {}}

    all_fn = []
    all_fp = []

    for root, dirs, files in os.walk(in_path):

        for f in files:
            # TODO: come up with something better to ensure one to one txt file comparisons with anno_path
            if not f.endswith(".txt"):
                continue

            # local values per file
            false_positives = []  # non-phi we think are phi
            false_positives_coords = []
            true_positives = []  # phi we correctly identify
            true_positives_coords = []
            false_negatives = []  # phi we think are non-phi
            false_negatives_coords = []
            true_negatives = []  # non-phi we correctly identify
            true_negatives_coords = []

            philtered_filename = root + f
            anno_filename = anno_path + "".join(f.split(".")[0]) + anno_suffix

            if not os.path.exists(philtered_filename):
                raise Exception("FILE DOESNT EXIST", philtered_filename)

            if not os.path.exists(anno_filename):
                continue

            philtered = open(philtered_filename, "r").read()
            philtered_words = re.split("(\s+)", philtered)
            philtered_words_cleaned = []
            for item in philtered_words:
                if len(item) > 0:
                    if item.isspace() == False:
                        split_item = re.split(
                            "(\s+)", re.sub(punctuation_matcher, " ", item)
                        )
                        for elem in split_item:
                            if len(elem) > 0:
                                philtered_words_cleaned.append(elem)
                    else:
                        philtered_words_cleaned.append(item)

            anno = open(anno_filename, "r").read()

            anno_words = re.split("(\s+)", anno)
            anno_words_cleaned = []
            for item in anno_words:
                if len(item) > 0:
                    if not item.isspace():
                        split_item = re.split(
                            "(\s+)", re.sub(punctuation_matcher, " ", item)
                        )
                        for elem in split_item:
                            if len(elem) > 0:
                                anno_words_cleaned.append(elem)
                    else:
                        anno_words_cleaned.append(item)
            for c, w, r in self.seq_eval(philtered_words_cleaned, anno_words_cleaned):

                # Double check that we aren't adding blank spaces or single punctionation characters to our lists
                if not w.isspace() and (re.sub(r"[^a-zA-Z0-9\*]+", "", w) != ""):

                    if c == "FP":
                        false_positives.append(w)
                        false_positives_coords.append([w, r])

                    elif c == "FN":
                        false_negatives.append(w)
                        false_negatives_coords.append([w, r])
                    elif c == "TP":
                        true_positives.append(w)
                        true_positives_coords.append([w, r])
                    elif c == "TN":
                        true_negatives.append(w)
                        true_negatives_coords.append([w, r])

            # update summary
            summary["summary_by_file"][philtered_filename] = {
                "false_positives": false_positives,
                "false_negatives": false_negatives,
                "num_false_negatives": len(false_negatives),
            }
            summary["total_true_positives"] = summary["total_true_positives"] + len(
                true_positives
            )
            summary["total_false_positives"] = summary["total_false_positives"] + len(
                false_positives
            )
            summary["total_false_negatives"] = summary["total_false_negatives"] + len(
                false_negatives
            )
            summary["total_true_negatives"] = summary["total_true_negatives"] + len(
                true_negatives
            )
            all_fp = all_fp + false_positives
            all_fn = all_fn + false_negatives

            # Create coordinate summaries
            summary_coords["summary_by_file"][philtered_filename] = {
                "false_positives": false_positives_coords,
                "false_negatives": false_negatives_coords,
                "true_positives": true_positives_coords,
            }

    if summary["total_true_positives"] + summary["total_false_negatives"] > 0:
        recall = summary["total_true_positives"] / (
            summary["total_true_positives"] + summary["total_false_negatives"]
        )
    elif summary["total_false_negatives"] == 0:
        recall = 1.0

    if summary["total_true_positives"] + summary["total_false_positives"] > 0:
        precision = summary["total_true_positives"] / (
            summary["total_true_positives"] + summary["total_false_positives"]
        )
    elif summary["total_true_positives"] == 0:
        precision = 0.0

    if summary["total_true_negatives"] + summary["total_false_positives"] > 0:
        retention = summary["total_true_negatives"] / (
            summary["total_true_negatives"] + summary["total_false_positives"]
        )
    else:
        retention = 0.0

    # DETAILED EVAL ##################
    # save the phi we missed
    json.dump(summary, open(summary_output, "w"), indent=4)
    json.dump(all_fn, open(fn_output, "w"), indent=4)
    json.dump(all_fp, open(fp_output, "w"), indent=4)

    if self.verbose:
        print("\n")
        print("Uncorrected Results:")
        print("\n")
        print(
            "TP:",
            summary["total_true_positives"],
            "FN:",
            summary["total_false_negatives"],
            "TN:",
            summary["total_true_negatives"],
            "FP:",
            summary["total_false_positives"],
        )
        print("Global Recall: {:.2%}".format(recall))
        print("Global Precision: {:.2%}".format(precision))
        print("Global Retention: {:.2%}".format(retention))

    # Get phi tags #####################
    # Get xml summary
    phi = self.xml
    # Create dictionary to hold fn tags
    fn_tags = {}
    fp_tags = {}

    # Keep track of recall and precision for each category
    phi_categories = ["Age", "Contact", "Date", "ID", "Location", "Name", "Other"]
    # i2b2:
    if not self.ucsf_format:
        # Define tag list
        i2b2_tags = [
            "DOCTOR",
            "PATIENT",
            "DATE",
            "MEDICALRECORD",
            "IDNUM",
            "DEVICE",
            "USERNAME",
            "PHONE",
            "EMAIL",
            "FAX",
            "CITY",
            "STATE",
            "ZIP",
            "STREET",
            "LOCATION-OTHER",
            "HOSPITAL",
            "AGE",
        ]

        i2b2_category_dict = {
            "DOCTOR": "Name",
            "PATIENT": "Name",
            "DATE": "Date",
            "MEDICALRECORD": "ID",
            "IDNUM": "ID",
            "DEVICE": "ID",
            "USERNAME": "Contact",
            "PHONE": "Contact",
            "EMAIL": "Contact",
            "FAX": "Contact",
            "CITY": "Location",
            "STATE": "Location",
            "ZIP": "Location",
            "STREET": "Location",
            "LOCATION-OTHER": "Location",
            "HOSPITAL": "Location",
            "AGE": "Age",
        }

        i2b2_include_tags = [
            "DOCTOR",
            "PATIENT",
            "DATE",
            "MEDICALRECORD",
            "IDNUM",
            "DEVICE",
            "USERNAME",
            "PHONE",
            "EMAIL",
            "FAX",
            "CITY",
            "STATE",
            "ZIP",
            "STREET",
            "LOCATION-OTHER",
            "HOSPITAL",
            "AGE",
        ]
        i2b2_patient_tags = [
            "PATIENT",
            "DATE",
            "MEDICALRECORD",
            "IDNUM",
            "DEVICE",
            "USERNAME",
            "PHONE",
            "EMAIL",
            "FAX",
            "CITY",
            "STATE",
            "ZIP",
            "STREET",
            "LOCATION-OTHER",
            "HOSPITAL",
            "AGE",
        ]
        i2b2_provider_tags = [
            "DOCTOR",
            "DATE",
            "USERNAME",
            "PHONE",
            "EMAIL",
            "FAX",
            "CITY",
            "STATE",
            "ZIP",
            "STREET",
            "LOCATION-OTHER",
            "HOSPITAL",
        ]

        rp_summaries = {}
        for i in range(0, len(i2b2_tags)):
            tag = i2b2_tags[i]
            fn_key = tag + "_fns"
            tp_key = tag + "_tps"
            rp_summaries[fn_key] = 0
            rp_summaries[tp_key] = 0

    # ucsf:
    if self.ucsf_format:
        # Define tag list
        ucsf_tags = [
            "Date",
            "Provider_Name",
            "Phone_Fax",
            "Age",
            "Patient_Name_or_Family_Member_Name",
            "Patient_Address",
            "Patient_Initials",
            "Provider_Address_or_Location",
            "Provider_Initials",
            "Provider_Certificate_or_License",
            "Patient_Medical_Record_Id",
            "Patient_Account_Number",
            "Patient_Social_Security_Number",
            "Patient_Vehicle_or_Device_Id",
            "Patient_Unique_Id",
            "Diagnosis_Code_ICD_or_International",
            "Procedure_or_Billing_Code",
            "Medical_Department_Name",
            "Email",
            "URL_IP",
            "Patient_Biometric_Id_or_Face_Photo",
            "Patient_Language_Spoken",
            "Patient_Place_Of_Work_or_Occupation",
            "Patient_Certificate_or_License",
            "Medical_Research_Study_Name_or_Number",
            "Teaching_Institution_Name",
            "Non_UCSF_Medical_Institution_Name",
            "Medical_Institution_Abbreviation",
            "Unclear",
        ]

        ucsf_category_dict = {
            "Date": "Date",
            "Provider_Name": "Name",
            "Phone_Fax": "Contact",
            "Age": "Age",
            "Patient_Name_or_Family_Member_Name": "Name",
            "Patient_Address": "Location",
            "Patient_Initials": "Name",
            "Provider_Address_or_Location": "Location",
            "Provider_Initials": "Name",
            "Provider_Certificate_or_License": "ID",
            "Patient_Medical_Record_Id": "ID",
            "Patient_Account_Number": "ID",
            "Patient_Social_Security_Number": "ID",
            "Patient_Vehicle_or_Device_Id": "ID",
            "Patient_Unique_Id": "ID",
            "Diagnosis_Code_ICD_or_International": "ID",
            "Procedure_or_Billing_Code": "ID",
            "Medical_Department_Name": "Location",
            "Email": "Contact",
            "URL_IP": "Contact",
            "Patient_Biometric_Id_or_Face_Photo": "ID",
            "Patient_Language_Spoken": "Other",
            "Patient_Place_Of_Work_or_Occupation": "Location",
            "Patient_Certificate_or_License": "ID",
            "Medical_Research_Study_Name_or_Number": "ID",
            "Teaching_Institution_Name": "Location",
            "Non_UCSF_Medical_Institution_Name": "Location",
            "Medical_Institution_Abbreviation": "Location",
            "Unclear": "Other",
        }

        if self.initials:
            ucsf_include_tags = [
                "Date",
                "Provider_Name",
                "Phone_Fax",
                "Patient_Name_or_Family_Member_Name",
                "Patient_Address",
                "Provider_Address_or_Location",
                "Provider_Certificate_or_License",
                "Patient_Medical_Record_Id",
                "Patient_Account_Number",
                "Patient_Social_Security_Number",
                "Patient_Vehicle_or_Device_Id",
                "Patient_Unique_Id",
                "Procedure_or_Billing_Code",
                "Email",
                "URL_IP",
                "Patient_Biometric_Id_or_Face_Photo",
                "Patient_Certificate_or_License",
                "Age",
                "Patient_Initials",
                "Provider_Initials",
            ]
            ucsf_patient_tags = [
                "Date",
                "Phone_Fax",
                "Age",
                "Patient_Name_or_Family_Member_Name",
                "Patient_Address",
                "Patient_Initials",
                "Patient_Medical_Record_Id",
                "Patient_Account_Number",
                "Patient_Social_Security_Number",
                "Patient_Vehicle_or_Device_Id",
                "Patient_Unique_Id",
                "Email",
                "URL_IP",
                "Patient_Biometric_Id_or_Face_Photo",
                "Patient_Certificate_or_License",
            ]
            ucsf_provider_tags = [
                "Provider_Name",
                "Phone_Fax",
                "Provider_Address_or_Location",
                "Provider_Initials",
                "Provider_Certificate_or_License",
                "Email",
                "URL_IP",
            ]

        else:
            ucsf_include_tags = [
                "Date",
                "Provider_Name",
                "Phone_Fax",
                "Patient_Name_or_Family_Member_Name",
                "Patient_Address",
                "Provider_Address_or_Location",
                "Provider_Certificate_or_License",
                "Patient_Medical_Record_Id",
                "Patient_Account_Number",
                "Patient_Social_Security_Number",
                "Patient_Vehicle_or_Device_Id",
                "Patient_Unique_Id",
                "Procedure_or_Billing_Code",
                "Email",
                "URL_IP",
                "Patient_Biometric_Id_or_Face_Photo",
                "Patient_Certificate_or_License",
                "Age",
            ]
            ucsf_patient_tags = [
                "Date",
                "Phone_Fax",
                "Age",
                "Patient_Name_or_Family_Member_Name",
                "Patient_Address",
                "Patient_Medical_Record_Id",
                "Patient_Account_Number",
                "Patient_Social_Security_Number",
                "Patient_Vehicle_or_Device_Id",
                "Patient_Unique_Id",
                "Email",
                "URL_IP",
                "Patient_Biometric_Id_or_Face_Photo",
                "Patient_Certificate_or_License",
            ]
            ucsf_provider_tags = [
                "Provider_Name",
                "Phone_Fax",
                "Provider_Address_or_Location",
                "Provider_Certificate_or_License",
                "Email",
                "URL_IP",
            ]

        rp_summaries = {}
        for i in range(0, len(ucsf_tags)):
            tag = ucsf_tags[i]
            fn_key = tag + "_fns"
            tp_key = tag + "_tps"
            rp_summaries[fn_key] = 0
            rp_summaries[tp_key] = 0

    # Create dictionaries for unigram and bigram PHI/non-PHI frequencies
    # Diciontary values look like: [phi_count, non-phi_count]
    unigram_dict = {}
    bigram_dict = {}
    corrected_age_fns = 0

    # Loop through all filenames in summary
    for fn in summary_coords["summary_by_file"]:
        # print(self.patterns)
        # get input notes filename (for filter analysis wit coordinatemap)
        input_filename = os.path.basename(fn)

        current_summary = summary_coords["summary_by_file"][fn]

        # Get corresponding info in phi_notes
        note_name = fn.split("/")[-1]

        try:
            anno_name = note_name.split(".")[0] + ".xml"
            text = phi[anno_name]["text"]
        except KeyError:
            anno_name = note_name.split(".")[0] + ".txt.xml"
            text = phi[anno_name]["text"]

        lst = re.split(r"(\s+)", text)
        cleaned = []
        for item in lst:
            if len(item) > 0:
                if not item.isspace():
                    split_item = re.split(r"(\s+)", re.sub(r"[^a-zA-Z0-9]", " ", item))
                    for elem in split_item:
                        if len(elem) > 0:
                            cleaned.append(elem)
                else:
                    cleaned.append(item)

        # Get coords for POS tags
        start_coordinate = 0
        pos_coords = []
        for item in cleaned:
            pos_coords.append(start_coordinate)
            start_coordinate += len(item)

        pos_list = nltk.pos_tag(cleaned)

        cleaned_with_pos = {}
        for i in range(0, len(pos_list)):
            cleaned_with_pos[str(pos_coords[i])] = [pos_list[i][0], pos_list[i][1]]

        # Get FN tags ##########
        phi_list = phi[anno_name]["phi"]

        # Create unigram and bigram frequency tables #######
        if self.freq_table:

            # Create separate cleaned list/coord list without spaces
            cleaned_nospaces = []
            coords_nospaces = []
            for i in range(0, len(cleaned)):
                if not cleaned[i].isspace():
                    cleaned_nospaces.append(cleaned[i])
                    coords_nospaces.append(pos_coords[i])

            # Loop through all single words and word pairs, and compare with PHI list
            for i in range(0, len(cleaned_nospaces) - 1):
                unigram_word = (
                    cleaned_nospaces[i]
                    .replace("\n", "")
                    .replace("\t", "")
                    .replace(" ", "")
                    .lower()
                )
                bigram_word = " ".join(
                    [
                        cleaned_nospaces[i]
                        .replace("\n", "")
                        .replace("\t", "")
                        .replace(" ", "")
                        .lower(),
                        cleaned_nospaces[i + 1]
                        .replace("\n", "")
                        .replace("\t", "")
                        .replace(" ", "")
                        .lower(),
                    ]
                )
                unigram_start = coords_nospaces[i]
                bigram_start1 = coords_nospaces[i]
                bigram_start2 = coords_nospaces[i + 1]

                # Loop through PHI list and compare ranges
                for phi_item in phi_list:
                    try:
                        phi_start = phi_item["start"]
                        phi_end = phi_item["end"]
                    except KeyError:
                        phi_start = phi_item["spans"].split("~")[0]
                        phi_end = phi_item["spans"].split("~")[1]
                    if unigram_start in range(int(phi_start), int(phi_end)):
                        # This word is PHI and hasn't been added to the dictionary yet
                        if unigram_word not in unigram_dict:
                            unigram_dict[unigram_word] = [1, 0]
                        # This word is PHI and has already been added to the dictionary
                        else:
                            unigram_dict[unigram_word][0] += 1
                    else:
                        # This word is not PHI and hasn't been aded to the dictionary yet
                        if unigram_word not in unigram_dict:
                            unigram_dict[unigram_word] = [0, 1]
                        # This word is not PHI and has already been added to the dictionary
                        else:
                            unigram_dict[unigram_word][1] += 1
                    if bigram_start1 in range(
                        int(phi_start), int(phi_end)
                    ) and bigram_start2 in range(int(phi_start), int(phi_end)):
                        # This word is PHI and hasn't been added to the dictionary yet
                        if bigram_word not in bigram_dict:
                            bigram_dict[bigram_word] = [1, 0]
                        # This word is PHI and has already been added to the dictionary
                        else:
                            bigram_dict[bigram_word][0] += 1
                    else:
                        # This word is not PHI and hasn't been aded to the dictionary yet
                        if bigram_word not in bigram_dict:
                            bigram_dict[bigram_word] = [0, 1]
                        # This word is not PHI and has already been added to the dictionary
                        else:
                            bigram_dict[bigram_word][1] += 1

        # Get tp counts per category
        current_tps = current_summary["true_positives"]
        # Initialize list to keep track of non-include tag FPs
        additional_fps = []

        for word in current_tps:
            start_coordinate_tp = word[1]
            for phi_item in phi_list:
                if self.ucsf_format:
                    phi_start = int(phi_item["spans"].split("~")[0])
                    phi_end = int(phi_item["spans"].split("~")[1])
                else:
                    phi_start = phi_item["start"]
                    phi_end = phi_item["end"]
                phi_type = phi_item["TYPE"]

                if not self.ucsf_format:
                    for i in range(0, len(i2b2_tags)):
                        tag = i2b2_tags[i]
                        tp_key = tag + "_tps"
                        if (
                            start_coordinate_tp in range(int(phi_start), int(phi_end))
                        ) and (tag == phi_type):
                            rp_summaries[tp_key] += 1
                    # Add these TPs to the FPs list of they are not in the include list
                    if phi_type not in i2b2_include_tags:
                        if start_coordinate_tp in range(int(phi_start), int(phi_end)):
                            additional_fps.append(
                                [
                                    text[
                                        start_coordinate_tp : start_coordinate_tp
                                        + len(word[0])
                                    ],
                                    start_coordinate_tp,
                                ]
                            )
                # ucsf
                if self.ucsf_format:
                    if phi_type not in ucsf_include_tags:
                        if start_coordinate_tp in range(int(phi_start), int(phi_end)):
                            additional_fps.append(
                                [
                                    text[
                                        start_coordinate_tp : start_coordinate_tp
                                        + len(word[0])
                                    ],
                                    start_coordinate_tp,
                                ]
                            )

                    for i in range(0, len(ucsf_tags)):
                        tag = ucsf_tags[i]
                        tp_key = tag + "_tps"
                        if (
                            start_coordinate_tp in range(int(phi_start), int(phi_end))
                        ) and (tag == phi_type):
                            rp_summaries[tp_key] += 1

        # i2b2
        if not self.ucsf_format:
            fn_counter_dict = {}
            for i in range(0, len(i2b2_tags)):
                tag = i2b2_tags[i]
                tag_fn_counter = tag + "_fn_counter"
                fn_counter_dict[tag_fn_counter] = 0
        # ucsf
        if self.ucsf_format:
            fn_counter_dict = {}
            for i in range(0, len(ucsf_tags)):
                tag = ucsf_tags[i]
                tag_fn_counter = tag + "_fn_counter"
                fn_counter_dict[tag_fn_counter] = 0

        fn_tag_summary = {}
        include_exclude_fns = ""

        if current_summary["false_negatives"] != [] and current_summary[
            "false_negatives"
        ] != [""]:
            counter = 0
            current_fns = current_summary["false_negatives"]

            for word in current_fns:
                counter += 1
                false_negative = word[0]
                start_coordinate_fn = word[1]

                # initialize list that will hold info on what matched what
                filter_file_list_exclude = []
                filter_file_list_include = []

                # if self.dependent:
                # Loop through coorinate map objects and match patterns with FPs
                for i, pattern in enumerate(self.patterns):
                    coord_map = pattern["coordinate_map"]
                    exclude_include = pattern["exclude"]
                    try:
                        filter_path = pattern["filepath"]
                    except KeyError:
                        filter_path = pattern["title"]
                    for start, stop in coord_map.filecoords(input_filename):
                        # Find intersection between ranges
                        word_range = set(
                            range(
                                start_coordinate_fn,
                                start_coordinate_fn + len(false_negative),
                            )
                        )
                        filter_range = set(range(start, stop))
                        intersection = word_range & filter_range
                        if intersection != set():
                            # Add this filter path to the list of things that filtered this word
                            if exclude_include:
                                filter_file_list_exclude.append(filter_path)
                            else:
                                filter_file_list_include.append(filter_path)

                for phi_item in phi_list:
                    phi_type = phi_item["TYPE"]
                    if self.ucsf_format:
                        phi_start = int(phi_item["spans"].split("~")[0])
                        phi_end = int(phi_item["spans"].split("~")[1])
                    else:
                        phi_start = phi_item["start"]
                        phi_end = phi_item["end"]

                    # i2b2
                    if not self.ucsf_format:
                        for i in range(0, len(i2b2_tags)):
                            tag = i2b2_tags[i]
                            fn_key = tag + "_fns"
                            tag_fn_counter = tag + "_fn_counter"
                            if (
                                start_coordinate_fn
                                in range(int(phi_start), int(phi_end))
                            ) and phi_type == tag:
                                rp_summaries[fn_key] += 1
                                fn_counter_dict[tag_fn_counter] += 1

                    # ucsf
                    if self.ucsf_format:
                        for i in range(0, len(ucsf_tags)):
                            tag = ucsf_tags[i]
                            fn_key = tag + "_fns"
                            tag_fn_counter = tag + "_fn_counter"
                            if (
                                start_coordinate_fn
                                in range(int(phi_start), int(phi_end))
                            ) and phi_type == tag:
                                if tag != "Age":
                                    rp_summaries[fn_key] += 1
                                    fn_counter_dict[tag_fn_counter] += 1

                    # Find PHI match: fn in text, coord in range
                    if start_coordinate_fn in range(int(phi_start), int(phi_end)):
                        # Get PHI tag
                        phi_tag = phi_type
                        # Get POS tag
                        pos_tag = cleaned_with_pos[str(start_coordinate_fn)][1]

                        # Get 25 characters surrounding FN on either side
                        context_start = start_coordinate_fn - 25
                        context_end = start_coordinate_fn + len(false_negative) + 25
                        if context_start >= 0 and context_end <= len(text) - 1:
                            fn_context = text[context_start:context_end]
                        elif context_start >= 0 and context_end > len(text) - 1:
                            fn_context = text[context_start:]
                        else:
                            fn_context = text[:context_end]

                        # Get fn id, to distinguish betweem multiple entries
                        fn_id = "N" + str(counter)

                        # Get include or exclude
                        if not self.ucsf_format:
                            if phi_tag in i2b2_include_tags:
                                include_exclude_fns = "include"
                            else:
                                include_exclude_fns = "exclude"
                        if self.ucsf_format:
                            if phi_tag in ucsf_include_tags:
                                if phi_tag != "Age":
                                    include_exclude_fns = "include"
                                # If age is over 90, include. Else, exclude
                                else:
                                    fn_stripped = false_negative.replace(".", "")
                                    # Is the age an integer?
                                    if fn_stripped.isdigit():
                                        if int(fn_stripped) >= 90:
                                            include_exclude_fns = "include"
                                            corrected_age_fns += 1
                                            # print('Include (int): ',fn_stripped)
                                        else:
                                            include_exclude_fns = "exclude"
                                    # Is the age a string?
                                    # Note that this won't catch all age FNs that are spelled ou
                                    # i.e., only the 'ninety' in 'ninety-five' will be marked as include
                                    # This won't affect our recall at all, but it will affect our precision a little
                                    # We will manually need to subtract theses from our FPs and add to our TPs
                                    else:
                                        if "ninety" in fn_stripped:
                                            include_exclude_fns = "include"
                                            corrected_age_fns += 1
                                        else:
                                            include_exclude_fns = "exclude"
                            else:
                                include_exclude_fns = "exclude"
                        # Create output dicitonary with id/word/pos/phi
                        fn_tag_summary[fn_id] = [
                            false_negative,
                            phi_tag,
                            pos_tag,
                            fn_context,
                            include_exclude_fns,
                            filter_file_list_exclude,
                            filter_file_list_include,
                        ]

        if fn_tag_summary != {}:
            fn_tags[fn] = fn_tag_summary

        # Get FP tags #########
        fp_tag_summary = {}
        current_fps = current_summary["false_positives"] + additional_fps
        if current_fps != [] and current_fps != [""]:
            counter = 0
            for word in current_fps:
                counter += 1
                false_positive = word[0]
                start_coordinate_fp = word[1]
                # print(word)

                # initialize list that will hold info on what matched what
                filter_file_list_exclude = []
                filter_file_list_include = []

                # if self.dependent:
                # Loop through coorinate map objects and match patterns with FPs
                for i, pattern in enumerate(self.patterns):
                    coord_map = pattern["coordinate_map"]
                    exclude_include = pattern["exclude"]
                    try:
                        filter_path = pattern["filepath"]
                    except KeyError:
                        filter_path = pattern["title"]
                    for start, stop in coord_map.filecoords(input_filename):
                        word_range = set(
                            range(
                                start_coordinate_fp,
                                start_coordinate_fp + len(false_positive),
                            )
                        )
                        filter_range = set(range(start, stop))
                        intersection = word_range & filter_range
                        if intersection != set():
                            # Add this filter path to the list of things that filtered this word
                            if exclude_include:
                                filter_file_list_exclude.append(filter_path)
                            else:
                                filter_file_list_include.append(filter_path)

                pos_entry = cleaned_with_pos[str(start_coordinate_fp)]

                pos_tag = pos_entry[1]

                # Get 25 characters surrounding FP on either side
                fp_context = ""
                context_start = start_coordinate_fp - 25
                context_end = start_coordinate_fp + len(false_positive) + 25
                if context_start >= 0 and context_end <= len(text) - 1:
                    fp_context = text[context_start:context_end]
                elif context_start >= 0 and context_end > len(text) - 1:
                    fp_context = text[context_start:]
                else:
                    fp_context = text[:context_end]

                fp_id = "P" + str(counter)

                fp_tag_summary[fp_id] = [
                    false_positive,
                    pos_tag,
                    fp_context,
                    filter_file_list_exclude,
                    filter_file_list_include,
                ]

        if fp_tag_summary != {}:
            fp_tags[fn] = fp_tag_summary

    # Create frequency table outputs
    if self.freq_table:
        # Unigram table
        with open("./data/phi/unigram_freq_table.csv", "w") as f:
            f.write("unigram,phi_count,non-phi_count\n")
            for key in unigram_dict:
                word = key
                phi_count = unigram_dict[key][0]
                non_phi_count = unigram_dict[key][1]
                f.write(word + "," + str(phi_count) + "," + str(non_phi_count) + "\n")
        with open("./data/phi/bigram_freq_table.csv", "w") as f:
            f.write("bigram,phi_count,non-phi_count\n")
            for key in bigram_dict:
                term = key
                phi_count = bigram_dict[key][0]
                non_phi_count = bigram_dict[key][1]
                f.write(term + "," + str(phi_count) + "," + str(non_phi_count) + "\n")

    # get specific recalls
    # i2b2
    overall_data = []
    if not self.ucsf_format:
        include_dict = {
            "fns": 0,
            "tps": 0,
            "fps": summary["total_false_positives"],
            "tns": summary["total_true_negatives"],
        }
        patient_phi_dict = {
            "fns": 0,
            "tps": 0,
            "fps": summary["total_false_positives"],
            "tns": summary["total_true_negatives"],
        }
        provider_phi_dict = {
            "fns": 0,
            "tps": 0,
            "fps": summary["total_false_positives"],
            "tns": summary["total_true_negatives"],
        }

        category_dict = {}
        for i in range(0, len(phi_categories)):
            category_tag = phi_categories[i]
            category_fns = category_tag + "_fns"
            category_tps = category_tag + "_tps"

            category_dict[category_fns] = 0
            category_dict[category_tps] = 0

        overall_recall_dict = {}

        for i in range(0, len(i2b2_tags)):
            tag = i2b2_tags[i]
            fn_key = tag + "_fns"
            tp_key = tag + "_tps"
            recall_key = tag + "_recall"

            # Get info for overall include dict and category dict
            if tag in i2b2_include_tags:
                include_dict["fns"] += rp_summaries[fn_key]
                include_dict["tps"] += rp_summaries[tp_key]

                if tag in i2b2_patient_tags:
                    patient_phi_dict["fns"] += rp_summaries[fn_key]
                    patient_phi_dict["tps"] += rp_summaries[tp_key]
                if tag in i2b2_provider_tags:
                    provider_phi_dict["fns"] += rp_summaries[fn_key]
                    provider_phi_dict["tps"] += rp_summaries[tp_key]

                tag_category = i2b2_category_dict[tag]
                category_fns = tag_category + "_fns"
                category_tps = tag_category + "_tps"

                category_dict[category_fns] += rp_summaries[fn_key]
                category_dict[category_tps] += rp_summaries[tp_key]

            # Get additional TNs and FPs
            if tag not in i2b2_include_tags:
                include_dict["tns"] += rp_summaries[fn_key]
                include_dict["fps"] += rp_summaries[tp_key]

                if tag in i2b2_patient_tags:
                    patient_phi_dict["tns"] += rp_summaries[fn_key]
                    patient_phi_dict["fps"] += rp_summaries[tp_key]
                if tag in i2b2_provider_tags:
                    provider_phi_dict["tns"] += rp_summaries[fn_key]
                    provider_phi_dict["fps"] += rp_summaries[tp_key]

            if rp_summaries[fn_key] != 0:
                overall_recall_dict[recall_key] = rp_summaries[tp_key] / (
                    rp_summaries[fn_key] + rp_summaries[tp_key]
                )
            else:
                overall_recall_dict[recall_key] = 1

            overall_data.append(
                [
                    tag,
                    "{:.2%}".format(overall_recall_dict[recall_key]),
                    str(rp_summaries[tp_key]),
                    str(rp_summaries[fn_key]),
                ]
            )

    # ucsf
    if self.ucsf_format:
        include_dict = {
            "fns": 0,
            "tps": 0,
            "fps": summary["total_false_positives"],
            "tns": summary["total_true_negatives"],
        }
        patient_phi_dict = {
            "fns": 0,
            "tps": 0,
            "fps": summary["total_false_positives"],
            "tns": summary["total_true_negatives"],
        }
        provider_phi_dict = {
            "fns": 0,
            "tps": 0,
            "fps": summary["total_false_positives"],
            "tns": summary["total_true_negatives"],
        }

        category_dict = {}
        for i in range(0, len(phi_categories)):
            category_tag = phi_categories[i]
            category_fns = category_tag + "_fns"
            category_tps = category_tag + "_tps"

            category_dict[category_fns] = 0
            category_dict[category_tps] = 0

        overall_recall_dict = {}

        for tag in ucsf_tags:
            fn_key = tag + "_fns"
            tp_key = tag + "_tps"
            recall_key = tag + "_recall"

            # Get info for overall include dict and category dict
            if tag in ucsf_include_tags:
                if tag != "Age":
                    include_dict["fns"] += rp_summaries[fn_key]
                    include_dict["tps"] += rp_summaries[tp_key]

                    if tag in ucsf_patient_tags:
                        patient_phi_dict["fns"] += rp_summaries[fn_key]
                        patient_phi_dict["tps"] += rp_summaries[tp_key]
                    if tag in ucsf_provider_tags:
                        provider_phi_dict["fns"] += rp_summaries[fn_key]
                        provider_phi_dict["tps"] += rp_summaries[tp_key]

                    tag_category = ucsf_category_dict[tag]
                    category_fns = tag_category + "_fns"
                    category_tps = tag_category + "_tps"

                    category_dict[category_fns] += rp_summaries[fn_key]
                    category_dict[category_tps] += rp_summaries[tp_key]
                else:
                    include_dict["fns"] += corrected_age_fns
                    include_dict["tps"] += rp_summaries[tp_key]
                    include_dict["tns"] += rp_summaries[fn_key] - corrected_age_fns

                    if tag in ucsf_patient_tags:
                        patient_phi_dict["fns"] += corrected_age_fns
                        patient_phi_dict["tps"] += rp_summaries[tp_key]
                        patient_phi_dict["tns"] += (
                            rp_summaries[fn_key] - corrected_age_fns
                        )
                    if tag in ucsf_provider_tags:
                        provider_phi_dict["fns"] += corrected_age_fns
                        provider_phi_dict["tps"] += rp_summaries[tp_key]
                        provider_phi_dict["tns"] += (
                            rp_summaries[fn_key] - corrected_age_fns
                        )

                    tag_category = ucsf_category_dict[tag]
                    category_fns = tag_category + "_fns"
                    category_tps = tag_category + "_tps"

                    category_dict[category_fns] += corrected_age_fns
                    category_dict[category_tps] += rp_summaries[tp_key]

            # Get additional TNs and FPs
            if tag not in ucsf_include_tags:
                include_dict["tns"] += rp_summaries[fn_key]
                include_dict["fps"] += rp_summaries[tp_key]

                if tag in ucsf_patient_tags:
                    patient_phi_dict["tns"] += rp_summaries[fn_key]
                    patient_phi_dict["fps"] += rp_summaries[tp_key]
                if tag in ucsf_provider_tags:
                    provider_phi_dict["tns"] += rp_summaries[fn_key]
                    provider_phi_dict["fps"] += rp_summaries[tp_key]

            if rp_summaries[fn_key] != 0:
                overall_recall_dict[recall_key] = rp_summaries[tp_key] / (
                    rp_summaries[fn_key] + rp_summaries[tp_key]
                )
            else:
                overall_recall_dict[recall_key] = 1
            if tag == "Age":
                overall_data.append(
                    [
                        tag,
                        "{:.2%}".format(overall_recall_dict[recall_key]),
                        str(rp_summaries[tp_key]),
                        str(corrected_age_fns),
                    ]
                )
            else:
                overall_data.append(
                    [
                        tag,
                        "{:.2%}".format(overall_recall_dict[recall_key]),
                        str(rp_summaries[tp_key]),
                        str(rp_summaries[fn_key]),
                    ]
                )

    # pretty print tag recalls
    overall_data.sort(key=lambda x: float(x[1][:-1]), reverse=True)
    sorted_overall_data = [["Tag", "Recall", "TPs", "FNs"]]
    for item in overall_data:
        sorted_overall_data.append(item)

    if self.verbose:
        print("\n")
        print("Recall by Tag:")
        col_width = (
            max(len(word) for row in sorted_overall_data for word in row) + 2
        )  # padding
        for row in sorted_overall_data:
            print("".join(word.ljust(col_width) for word in row))

    # Get category recall
    category_data = []
    for i in range(0, len(phi_categories)):
        category_tag = phi_categories[i]
        category_fns = category_tag + "_fns"
        category_tps = category_tag + "_tps"

        if category_dict[category_fns] != 0:
            category_recall = category_dict[category_tps] / (
                category_dict[category_fns] + category_dict[category_tps]
            )
        else:
            category_recall = 1
        category_data.append(
            [
                category_tag,
                "{:.2%}".format(category_recall),
                str(category_dict[category_tps]),
                str(category_dict[category_fns]),
            ]
        )

    # pretty print category recalls
    category_data.sort(key=lambda x: float(x[1][:-1]), reverse=True)
    sorted_category_data = [["Category", "Recall", "TPs", "FNs"]]
    for item in category_data:
        sorted_category_data.append(item)
    if self.verbose:
        print("\n")
        print("Recall by PHI Category:")
        col_width = (
            max(len(word) for row in category_data for word in row) + 2
        )  # padding
        for row in sorted_category_data:
            print("".join(word.ljust(col_width) for word in row))

    # Get corrected recall, precision ##########

    if include_dict["fns"] != 0:
        corrected_recall = include_dict["tps"] / (
            include_dict["fns"] + include_dict["tps"]
        )
    else:
        corrected_recall = 1

    if include_dict["fps"] != 0:
        corrected_precision = include_dict["tps"] / (
            include_dict["fps"] + include_dict["tps"]
        )
    else:
        corrected_precision = 1

    if include_dict["fps"] != 0:
        specificity = include_dict["tns"] / (include_dict["fps"] + include_dict["tns"])
    else:
        specificity = 1

    print("\n")
    print("Corrected Results:")
    print("\n")
    print(
        "cTP:",
        include_dict["tps"],
        "cFN:",
        include_dict["fns"],
        "cTN:",
        include_dict["tns"],
        "cFP:",
        include_dict["fps"],
    )
    print("Corrected Recall: " + "{:.2%}".format(corrected_recall))
    print("Corrected Precision: " + "{:.2%}".format(corrected_precision))
    print("Corrected Retention: " + "{:.2%}".format(specificity))
    print("\n")

    # Patient-only recall, precision ##########

    if patient_phi_dict["fns"] != 0:
        patient_recall = patient_phi_dict["tps"] / (
            patient_phi_dict["fns"] + patient_phi_dict["tps"]
        )
    else:
        patient_recall = 1

    if patient_phi_dict["fps"] != 0:
        patient_precision = patient_phi_dict["tps"] / (
            patient_phi_dict["fps"] + patient_phi_dict["tps"]
        )
    else:
        patient_precision = 1

    if patient_phi_dict["fps"] != 0:
        patient_specificity = patient_phi_dict["tns"] / (
            patient_phi_dict["fps"] + patient_phi_dict["tns"]
        )
    else:
        patient_specificity = 1

    print("\n")
    print("Patient-Only Results:")
    print("\n")
    print(
        "cTP:",
        patient_phi_dict["tps"],
        "cFN:",
        patient_phi_dict["fns"],
        "cTN:",
        patient_phi_dict["tns"],
        "cFP:",
        patient_phi_dict["fps"],
    )
    print("Patient PHI Recall: " + "{:.2%}".format(patient_recall))
    print("Precision: " + "{:.2%}".format(patient_precision))
    print("Retention: " + "{:.2%}".format(patient_specificity))
    print("\n")

    # Provider-only recall, precision ##########

    provider_recall = 0
    if provider_phi_dict["fns"] != 0:
        provider_recall = provider_phi_dict["tps"] / (
            provider_phi_dict["fns"] + provider_phi_dict["tps"]
        )

    provider_precision = 0
    if provider_phi_dict["fps"] != 0:
        provider_precision = provider_phi_dict["tps"] / (
            provider_phi_dict["fps"] + provider_phi_dict["tps"]
        )

    if provider_phi_dict["fps"] != 0:
        # if include_dict['tps'] != 0 and (include_dict['tps']-include_dict['fns']) > 0:
        provider_specificity = provider_phi_dict["tns"] / (
            provider_phi_dict["fps"] + provider_phi_dict["tns"]
        )
        # else:
        #     corrected_recall = 0
    else:
        provider_specificity = 1

    print("\n")
    print("Provider-Only Results:")
    print("\n")
    print(
        "cTP:",
        provider_phi_dict["tps"],
        "cFN:",
        provider_phi_dict["fns"],
        "cTN:",
        provider_phi_dict["tns"],
        "cFP:",
        provider_phi_dict["fps"],
    )
    print("Provider PHI Recall: " + "{:.2%}".format(provider_recall))
    print("Precision: " + "{:.2%}".format(provider_precision))
    print("Retention: " + "{:.2%}".format(provider_specificity))
    print("\n")

    # Summarize FN results #########

    # With and without context #####

    # With context:
    # Condensed tags will contain id, word, PHI tag, POS tag, occurrences
    fn_tags_condensed_context = {}
    # Stores lists that represent distinct groups of words, PHI and POS tags
    fn_tags_condensed_list_context = []

    # No context:
    # Condensed tags will contain id, word, PHI tag, POS tag, occurrences
    fn_tags_condensed = {}
    # Stores lists that represent distinct groups of words, PHI and POS tags
    fn_tags_condensed_list = []

    # Keep track of how many distinct combinations we've added to each list
    context_counter = 0
    nocontext_counter = 0
    for fn in fn_tags:
        file_dict = fn_tags[fn]
        for subfile in file_dict:
            current_list_context = file_dict[subfile]
            ##############################
            # print(current_list_context)
            current_list_nocontext = (
                current_list_context[:3]
                + [current_list_context[-3]]
                + [current_list_context[-2]]
                + [current_list_context[-1]]
            )
            ############################

            word = current_list_context[0]
            phi_tag = current_list_context[1]
            pos_tag = current_list_context[2]
            fn_context = current_list_context[3].replace("\n", " ")
            filter_matches_exclude = current_list_context[5]
            filter_matches_include = current_list_context[6]

            # Context: add each occurrence with corresponding filename
            fn_tags_condensed_list_context.append(current_list_context)
            key_name = "uniq" + str(context_counter)
            filename = fn.split("/")[-1]
            include_exclude = current_list_context[4]
            fn_tags_condensed_context[key_name] = [
                word,
                phi_tag,
                pos_tag,
                fn_context,
                filename,
                include_exclude,
                filter_matches_exclude,
                filter_matches_include,
            ]
            context_counter += 1

            # No context
            if current_list_nocontext not in fn_tags_condensed_list:
                fn_tags_condensed_list.append(current_list_nocontext)
                key_name = "uniq" + str(nocontext_counter)
                fn_tags_condensed[key_name] = [
                    word,
                    phi_tag,
                    pos_tag,
                    1,
                    include_exclude,
                    filter_matches_exclude,
                    filter_matches_include,
                ]
                nocontext_counter += 1
            else:
                uniq_id_index = fn_tags_condensed_list.index(current_list_nocontext)
                uniq_id = "uniq" + str(uniq_id_index)
                fn_tags_condensed[uniq_id][3] += 1

    # Summariz FP results

    # With context
    # Condensed tags will contain id, word, POS tag, occurrences
    fp_tags_condensed_context = {}
    # Stores lists that represent distinct groups of wordss and POS tags
    fp_tags_condensed_list_context = []

    # No context
    # Condensed tags will contain id, word, POS tag, occurrences
    fp_tags_condensed = {}
    # Stores lists that represent distinct groups of wordss and POS tags
    fp_tags_condensed_list = []

    # Keep track of how many distinct combinations we've added to each list
    nocontext_counter = 0
    context_counter = 0
    for fp in fp_tags:
        file_dict = fp_tags[fp]
        for subfile in file_dict:
            current_list_context = file_dict[subfile]
            current_list_nocontext = (
                current_list_context[:2]
                + [current_list_context[3]]
                + [current_list_context[4]]
            )

            word = current_list_context[0]
            pos_tag = current_list_context[1]
            fp_context = current_list_context[2].replace("\n", " ")
            filter_matches_exclude = current_list_context[3]
            filter_matches_include = current_list_context[4]

            # Context: add each occurrence with corresponding filename
            fp_tags_condensed_list_context.append(current_list_context)
            key_name = "uniq" + str(context_counter)
            filename = fp.split("/")[-1]
            fp_tags_condensed_context[key_name] = [
                word,
                pos_tag,
                fp_context,
                filename,
                filter_matches_exclude,
                filter_matches_include,
            ]
            context_counter += 1

            # No Context
            if current_list_nocontext not in fp_tags_condensed_list:
                fp_tags_condensed_list.append(current_list_nocontext)
                key_name = "uniq" + str(nocontext_counter)
                fp_tags_condensed[key_name] = [
                    word,
                    pos_tag,
                    1,
                    filter_matches_exclude,
                    filter_matches_include,
                ]
                nocontext_counter += 1
            else:
                uniq_id_index = fp_tags_condensed_list.index(current_list_nocontext)
                uniq_id = "uniq" + str(uniq_id_index)
                fp_tags_condensed[uniq_id][2] += 1

    # Write FN and FP results to outfolder
    # Conext
    with open(self.eval_outpath + "fn_tags_context.txt", "w") as fn_file:
        fn_file.write(
            "key"
            + "|"
            + "note_word"
            + "|"
            + "phi_tag"
            + "|"
            + "pos_tag"
            + "|"
            + "context"
            + "|"
            + "filename"
            + "|"
            + "include_exclude"
            + "|"
            + "exclude_filters"
            + "|"
            + "include_filters"
            + "\n"
        )
        # print(fn_tags_condensed_context)
        for key in fn_tags_condensed_context:
            current_list = fn_tags_condensed_context[key]
            fn_file.write(
                key
                + "|"
                + current_list[0]
                + "|"
                + current_list[1]
                + "|"
                + current_list[2]
                + "|"
                + current_list[3]
                + "|"
                + current_list[4]
                + "|"
                + current_list[5]
                + "|"
                + str(current_list[6])
                + "|"
                + str(current_list[7])
                + "\n"
            )

    with open(self.eval_outpath + "fp_tags_context.txt", "w") as fp_file:
        fp_file.write(
            "key"
            + "|"
            + "note_word"
            + "|"
            + "pos_tag"
            + "|"
            + "context"
            + "|"
            + "filename"
            + "|"
            + "exclude_filters"
            + "|"
            + "include_filters"
            + "\n"
        )
        for key in fp_tags_condensed_context:
            current_list = fp_tags_condensed_context[key]
            fp_file.write(
                key
                + "|"
                + current_list[0]
                + "|"
                + current_list[1]
                + "|"
                + current_list[2]
                + "|"
                + current_list[3]
                + "|"
                + str(current_list[4])
                + "|"
                + str(current_list[5])
                + "\n"
            )

    # No context
    with open(self.eval_outpath + "fn_tags.txt", "w") as fn_file:
        fn_file.write(
            "key"
            + "|"
            + "note_word"
            + "|"
            + "phi_tag"
            + "|"
            + "pos_tag"
            + "|"
            + "occurrences"
            + "|"
            + "include_exclude"
            + "|"
            + "exclude_filters"
            + "|"
            + "include_filters"
            + "\n"
        )
        for key in fn_tags_condensed:
            current_list = fn_tags_condensed[key]
            fn_file.write(
                key
                + "|"
                + current_list[0]
                + "|"
                + current_list[1]
                + "|"
                + current_list[2]
                + "|"
                + str(current_list[3])
                + "|"
                + current_list[4]
                + "|"
                + str(current_list[5])
                + "|"
                + str(current_list[6])
                + "\n"
            )

    with open(self.eval_outpath + "fp_tags.txt", "w") as fp_file:
        fp_file.write(
            "key"
            + "|"
            + "note_word"
            + "|"
            + "pos_tag"
            + "|"
            + "occurrences"
            + "|"
            + "exclude_filters"
            + "|"
            + "include_filters"
            + "\n"
        )
        for key in fp_tags_condensed:
            current_list = fp_tags_condensed[key]
            fp_file.write(
                key
                + "|"
                + current_list[0]
                + "|"
                + current_list[1]
                + "|"
                + str(current_list[2])
                + "|"
                + str(current_list[3])
                + "|"
                + str(current_list[4])
                + "\n"
            )


def getphi(
    self,
    anno_folder="data/i2b2_anno/",
    anno_suffix="_phi_reduced.ano",
    data_folder="data/i2b2_notes/",
    output_folder="i2b2_phi",
    filter_regex=None,
):
    """ get's phi from existing data to build up a data model
    data structure to hold our phi and classify phi we find
        {
            "foo.txt":[
                {
                    "phi":"1/1/2019",
                    "context":"The data was 1/1/2019 and the patient was happy",
                    "class":"numer" //number, string ...
                },...
            ],...
        }
    """
    if self.run_eval:
        print("getphi")

    # use config if exists
    if self.anno_folder is not None:
        anno_folder = self.anno_folder

    if self.anno_suffix != "":
        anno_suffix = self.anno_suffix

    phi = {}
    word_counts = {}
    not_phi = {}

    for root, dirs, files in os.walk(data_folder):

        for f in files:

            if not os.path.exists(root + f):
                raise Exception("FILE DOESNT EXIST", root + f)

            if len(anno_suffix) > 0:
                if not os.path.exists(anno_folder + f.split(".")[0] + anno_suffix):
                    print(
                        "FILE DOESNT EXIST",
                        anno_folder + f.split(".")[0] + anno_suffix,
                    )
                    continue
            else:
                if not os.path.exists(anno_folder + f):
                    print("FILE DOESNT EXIST", anno_folder + f)
                    continue

            orig_filename = root + f
            encoding1 = self.detect_encoding(orig_filename)
            orig = open(orig_filename, "r", encoding=encoding1["encoding"]).read()

            orig_words = re.split("\s+", orig)

            anno_filename = anno_folder + f.split(".")[0] + anno_suffix
            encoding2 = self.detect_encoding(anno_filename)
            anno = open(anno_filename, "r", encoding=encoding2["encoding"]).read()
            anno_words = re.split("\s+", anno)

            anno_dict = {}

            for w in anno_words:
                anno_dict[w] = 1

            for i, w in enumerate(orig_words):

                # check for edge cases that should not be "words"
                x = w.replace("_", "").strip()
                if len(x) == 0:
                    continue

                # add all words to our counts
                if w not in word_counts:
                    word_counts[w] = 0
                word_counts[w] += 1

                # check if this word is phi
                if w not in anno_dict:

                    left_index = i - 10
                    if left_index < 0:
                        left_index = 0

                    right_index = i + 10
                    if right_index >= len(orig_words):
                        right_index = len(orig_words) - 1
                    window = orig_words[left_index:right_index]
                    if f not in phi:
                        phi[f] = []

                    c = "string"
                    if re.search("\d+", w):
                        c = "number"

                    phi[f].append({"phi": w, "context": window, "class": c})
                else:
                    # add all words to our counts
                    if w not in not_phi:
                        not_phi[w] = 0
                    not_phi[w] += 1

    # save our phi with context
    json.dump(phi, open("data/phi/phi_context.json", "w"), indent=4)

    # save all phi word counts
    counts = {}
    num_phi = {}
    string_phi = {}

    for f in phi:
        for d in phi[f]:
            if d["phi"] not in counts:
                counts[d["phi"]] = 0
            counts[d["phi"]] += 1
            if d["class"] == "number":
                if d["phi"] not in num_phi:
                    num_phi[d["phi"]] = 0
                num_phi[d["phi"]] += 1
            else:
                if d["phi"] not in string_phi:
                    string_phi[d["phi"]] = 0
                string_phi[d["phi"]] += 1

    # save all phi counts
    json.dump(counts, open("data/phi/phi_counts.json", "w"), indent=4)
    # save phi number counts
    json.dump(num_phi, open("data/phi/phi_number_counts.json", "w"), indent=4)
    # save phi string counts
    json.dump(string_phi, open("data/phi/phi_string_counts.json", "w"), indent=4)
    # save our total word counts
    json.dump(word_counts, open("data/phi/word_counts.json", "w"), indent=4)
    # save our total non_phi counts
    json.dump(not_phi, open("data/phi/non_phi_counts.json", "w"), indent=4)

    # get all non_phi counts by number or string
    non_phi_number = {}
    non_phi_string = {}
    for w in word_counts:
        if re.search("\d+", w):
            if w not in non_phi_number:
                non_phi_number[w] = 0
            non_phi_number[w] += 1
        else:
            if w not in non_phi_string:
                non_phi_string[w] = 0
            non_phi_string[w] += 1

    # save all phi string counts
    json.dump(
        non_phi_number, open("data/phi/non_phi_number_counts.json", "w"), indent=4
    )

    # save all phi number counts
    json.dump(
        non_phi_string, open("data/phi/non_phi_string_counts.json", "w"), indent=4
    )


def mapphi(
    self,
    phi_path="data/phi/phi_counts.json",
    out_path="data/phi/phi_map.json",
    sorted_path="data/phi/phi_sorted.json",
    digit_char="`",
    string_char="?",
):
    """ given all examples of the phi, creates a general representation

        digit_char = this is what digits are replaced by
        string_char = this is what strings are replaced by
        any_char = this is what any random characters are replaced with
    """

    d = json.load(open(phi_path, "r"))

    phi_map = {}

    for phi in d:
        wordlst = []
        phi_word = phi["phi"]
        for c in phi_word:
            if re.match("\d+", c):
                wordlst.append(digit_char)
            elif re.match("[a-zA-Z]+", c):
                wordlst.append(string_char)
            else:
                wordlst.append(c)
        word = "".join(wordlst)
        if word not in phi_map:
            phi_map[word] = {"examples": {}}
        if phi_word not in phi_map[word]["examples"]:
            phi_map[word]["examples"][phi_word] = []
        phi_map[word]["examples"][phi_word].append(phi)

    # save the count of all representations
    for k in phi_map:
        phi_map[k]["count"] = len(phi_map[k]["examples"].keys())

    # save all representations
    json.dump(phi_map, open(out_path, "w"), indent=4)

    # save an ordered list of representations so we can prioritize regex building
    items = []
    for k in phi_map:
        items.append(
            {
                "pattern": k,
                "examples": phi_map[k]["examples"],
                "count": len(phi_map[k]["examples"].keys()),
            }
        )

    items.sort(key=lambda x: x["count"], reverse=True)
    json.dump(items, open(sorted_path, "w"), indent=4)



import json
import os
import pickle
import re
import subprocess
import warnings
from dataclasses import dataclass
from typing import Dict, List, Optional, Pattern

import nltk
import pkg_resources
from chardet.universaldetector import UniversalDetector
from nltk.tag.stanford import StanfordNERTagger

from philter_lite.coordinate_map import CoordinateMap

DEFAULT_PHI_TYPE_LIST = [
    "DATE",
    "Patient_Social_Security_Number",
    "Email",
    "Provider_Address_or_Location",
    "Age",
    "Name",
    "OTHER",
]


@dataclass(frozen=True)
class Filter:
    title: str
    type: str
    exclude: bool
    notes: str
    phi_type: Optional[str]


@dataclass(frozen=True)
class SetFilter(Filter):
    filepath: str
    pos: List[str]
    data: Pattern[str]


@dataclass(frozen=True)
class RegexFilter(Filter):
    filepath: str
    data: Pattern[str]


@dataclass(frozen=True)
class RegexContextFilter(Filter):
    context: str
    context_filter: str
    filepath: str
    data: Pattern[str]


@dataclass(frozen=True)
class PosFilter(Filter):
    pos: List[str]


@dataclass(frozen=True)
class NerFilter(Filter):
    pos: Optional[List[str]]


@dataclass(frozen=True)
class PhiEntry:
    start: int
    stop: int
    word: str
    phi_type: str


@dataclass(frozen=True)
class NonPhiEntry:
    start: int
    stop: int
    word: str
    filepath: str


@dataclass(frozen=True)
class DataTracker:
    text: str
    phi: List[PhiEntry]
    non_phi: List[NonPhiEntry]


def precompile(filepath):
    """ precompiles our regex to speed up pattern matching"""
    regex = open(filepath, "r").read().strip()
    # NOTE: this is not thread safe! but we want to print a more detailed warning message
    with warnings.catch_warnings():
        warnings.simplefilter(
            action="error", category=FutureWarning
        )  # in order to print a detailed message
        try:
            re_compiled = re.compile(regex)
        except FutureWarning as warn:
            print("FutureWarning: {0} in file ".format(warn) + filepath)
            warnings.simplefilter(action="ignore", category=FutureWarning)
            re_compiled = re.compile(regex)  # assign nevertheless
    return re_compiled


def init_set(filepath):
    """ loads a set of words, (must be a dictionary or set shape) returns result"""
    if filepath.endswith(".pkl"):
        try:
            with open(filepath, "rb") as pickle_file:
                set_data = pickle.load(pickle_file)
        except UnicodeDecodeError:
            with open(filepath, "rb") as pickle_file:
                set_data = pickle.load(pickle_file, encoding="latin1")
    elif filepath.endswith(".json"):
        set_data = json.loads(open(filepath, "r").read())

    else:
        raise Exception("Invalid filteype", filepath)
    return set_data


def filter_from_dict(filter_dict):
    known_pattern_types = {
        "regex",
        "set",
        "regex_context",
        "stanford_ner",
        "pos_matcher",
        "match_all",
    }
    set_filetypes = {"pkl", "json"}
    regex_filetypes = {"txt"}

    filter_type = filter_dict["type"]

    resource_package = __name__

    if filter_type not in known_pattern_types:
        raise Exception("Pattern type is unknown", filter_type)

    if filter_type == "set":
        filter_path = pkg_resources.resource_filename(
            resource_package, filter_dict["filepath"]
        )
        if not os.path.exists(filter_path):
            raise Exception("Config filepath does not exist", filter_path)
        if filter_path.split(".")[-1] not in set_filetypes:
            raise Exception(
                "Invalid filteype", filter_path, "must be of", set_filetypes
            )
        data = init_set(filter_path)
        return SetFilter(
            title=filter_dict["title"],
            type=filter_type,
            exclude=filter_dict["exclude"],
            filepath=filter_path,
            notes=filter_dict["notes"],
            data=data,
            pos=filter_dict["pos"],
            phi_type=filter_dict.get("phi_type", None),
        )
    elif filter_type == "regex":
        filter_path = pkg_resources.resource_filename(
            resource_package, filter_dict["filepath"]
        )

        if not os.path.exists(filter_path):
            raise Exception("Config filepath does not exist", filter_path)
        if filter_path.split(".")[-1] not in regex_filetypes:
            raise Exception(
                "Invalid filteype", filter_path, "must be of", regex_filetypes
            )
        data = precompile(filter_path)
        return RegexFilter(
            title=filter_dict["title"],
            type=filter_type,
            exclude=filter_dict["exclude"],
            filepath=filter_path,
            notes=filter_dict["notes"],
            data=data,
            phi_type=filter_dict.get("phi_type", None),
        )

    elif filter_type == "regex_context":
        filter_path = pkg_resources.resource_filename(
            resource_package, filter_dict["filepath"]
        )
        if not os.path.exists(filter_path):
            raise Exception("Config filepath does not exist", filter_path)
        if filter_path.split(".")[-1] not in regex_filetypes:
            raise Exception(
                "Invalid filtertype", filter_path, "must be of", regex_filetypes
            )
        data = precompile(filter_path)

        return RegexContextFilter(
            title=filter_dict["title"],
            type=filter_type,
            exclude=filter_dict["exclude"],
            filepath=filter_dict["filepath"],
            notes=filter_dict["notes"],
            context=filter_dict["context"],
            context_filter=filter_dict["context_filter"],
            data=data,
            phi_type=filter_dict.get("phi_type", None),
        )
    elif filter_type == "pos_matcher":
        return PosFilter(
            title=filter_dict["title"],
            type=filter_type,
            exclude=filter_dict["exclude"],
            notes=filter_dict["notes"],
            pos=filter_dict["pos"],
            phi_type=filter_dict.get("phi_type", None),
        )
    else:
        return Filter(
            title=filter_dict["title"],
            type=filter_type,
            exclude=filter_dict["exclude"],
            notes=filter_dict["notes"],
            phi_type=filter_dict.get("phi_type", None),
        )


def build_filters(filter_path) -> List[Filter]:
    if not os.path.exists(filter_path):
        raise Exception("Filepath does not exist", filter_path)
    with open(filter_path, "r") as fil_file:
        return [filter_from_dict(x) for x in json.loads(fil_file.read())]


def build_ner_tagger(
    classifier, tagger_jar, download: bool = True
) -> StanfordNERTagger:
    if not os.path.exists(classifier) and not download:
        raise Exception(
            "Filepath does not exist", classifier,
        )
    else:
        # download the ner data
        process = subprocess.Popen(
            "cd generate_dataset && ./download_ner.sh".split(), stdout=subprocess.PIPE,
        )
        process.communicate()

    if not os.path.exists(tagger_jar):
        raise Exception("Filepath does not exist", tagger_jar)

    return StanfordNERTagger(classifier, tagger_jar)


def get_pos(cleaned):
    return nltk.pos_tag(cleaned)


def get_clean(text, pre_process=r"[^a-zA-Z0-9]"):
    # Use pre-process to split sentence by spaces AND symbols, while preserving spaces in the split list
    lst = re.split(r"(\s+)", text)
    cleaned = []
    for item in lst:
        if len(item) > 0:
            if not item.isspace():
                split_item = re.split(r"(\s+)", re.sub(pre_process, " ", item))
                for elem in split_item:
                    if len(elem) > 0:
                        cleaned.append(elem)
            else:
                cleaned.append(item)
    return cleaned


def map_coordinates(
    text_data: str,
    patterns: List[Filter],
    phi_type_list: List[str] = DEFAULT_PHI_TYPE_LIST,
):
    """ Runs the set, or regex on the input data
        generating a coordinate map of hits given
        (this performs a dry run on the data and doesn't transform)
    """
    # create coordinate maps for each pattern
    pattern_coords = {}
    for pat in patterns:
        pattern_coords[pat.title] = CoordinateMap()

    # Get full self.include/exclude map before transform
    data_tracker = DataTracker(text_data, [], [])

    # create an intersection map of all coordinates we'll be removing
    exclude_map = CoordinateMap()

    # create an interestion map of all coordinates we'll be keeping
    include_map = CoordinateMap()

    # add file to phi_type_dict
    phi_type_dict = {}
    for phi_type in phi_type_list:
        phi_type_dict[phi_type] = CoordinateMap()

    # Create inital self.exclude/include for file
    for i, pat in enumerate(patterns):
        pattern_coord = pattern_coords[pat.title]

        if pat.type == "regex" and isinstance(pat, RegexFilter):
            map_regex(text=text_data, coord_map=pattern_coord, pattern=pat)
        elif pat.type == "set" and isinstance(pat, SetFilter):
            map_set(text=text_data, coord_map=pattern_coord, pattern=pat)
        elif pat.type == "regex_context" and isinstance(pat, RegexContextFilter):
            map_regex_context(
                text=text_data,
                coord_map=pattern_coord,
                all_patterns=pattern_coords,
                include_map=include_map,
                pattern=pat,
            )
        elif pat.type == "stanford_ner" and isinstance(pat, NerFilter):
            map_ner(text=text_data, pattern=pat)
        elif pat.type == "pos_matcher" and isinstance(pat, PosFilter):
            map_pos(text=text_data, coord_map=pattern_coord, pattern=pat)
        elif pat.type == "match_all":
            match_all(text=text_data, coord_map=pattern_coord)
        else:
            raise Exception("Error, pattern type not supported: ", pat.type)
        get_exclude_include_maps(
            pat,
            text_data,
            pattern_coord,
            include_map,
            exclude_map,
            phi_type_dict,
            data_tracker,
        )

    # create intersection maps for all phi types and add them to a dictionary containing all maps
    # get full exclude map (only updated either on-command by map_regex_context or at the very end of map_
    # coordinates)
    full_exclude_map = include_map.get_complement(text_data)

    for phi_type in phi_type_list:
        for start, stop in phi_type_dict[phi_type].filecoords():
            data_tracker.phi.append(
                PhiEntry(
                    start=start,
                    stop=stop,
                    word=text_data[start:stop],
                    phi_type=phi_type,
                )
            )

    return text_data, include_map, exclude_map, data_tracker


def map_regex(
    text, pattern: RegexFilter, coord_map: CoordinateMap, pre_process=r"[^a-zA-Z0-9]",
) -> CoordinateMap:
    """ Creates a coordinate map from the pattern on this data
        generating a coordinate map of hits given (dry run doesn't transform)
    """
    regex = pattern.data

    # All regexes except matchall
    if regex != re.compile("."):
        matches = regex.finditer(text)

        for m in matches:
            coord_map.add_extend(m.start(), m.start() + len(m.group()))

        return coord_map

    # MATCHALL/CATCHALL
    elif regex == re.compile("."):
        # Split note the same way we would split for set or POS matching
        matchall_list = re.split(r"(\s+)", text)
        matchall_list_cleaned = []
        for item in matchall_list:
            if len(item) > 0:
                if not item.isspace():
                    split_item = re.split(r"(\s+)", re.sub(pre_process, " ", item))
                    for elem in split_item:
                        if len(elem) > 0:
                            matchall_list_cleaned.append(elem)
                else:
                    matchall_list_cleaned.append(item)

        start_coordinate = 0
        for word in matchall_list_cleaned:
            start = start_coordinate
            stop = start_coordinate + len(word)
            word_clean = re.sub(r"[^a-zA-Z0-9]+", "", word.lower().strip())
            if len(word_clean) == 0:
                # got a blank space or something without any characters or digits, move forward
                start_coordinate += len(word)
                continue

            if regex.match(word_clean):
                coord_map.add_extend(start, stop)

            # advance our start coordinate
            start_coordinate += len(word)

        return coord_map


def map_regex_context(
    text,
    pattern: RegexContextFilter,
    coord_map: CoordinateMap,
    all_patterns: Dict[str, CoordinateMap],
    include_map: CoordinateMap,
    pre_process=r"[^a-zA-Z0-9]",
) -> CoordinateMap:
    """ map_regex_context creates a coordinate map from combined regex + PHI coordinates
    of all previously mapped patterns
    """
    regex = pattern.data
    context = pattern.context
    try:
        context_filter = pattern.context_filter
    except KeyError:
        warnings.warn(
            f"deprecated missing context_filter field in filter {pattern.title} of "
            f"type regex_context, assuming 'all'",
            DeprecationWarning,
        )
        context_filter = "all"

    # Get PHI coordinates
    if context_filter == "all":
        current_include_map = include_map
        # Create complement exclude map (also excludes punctuation)
        full_exclude_map = current_include_map.get_complement(text)

    else:
        full_exclude_map_coordinates = all_patterns[context_filter]
        full_exclude_map = {}
        for start, stop in full_exclude_map_coordinates.filecoords():
            full_exclude_map[start] = stop

    # 1. Get coordinates of all include and exclude mathches

    punctuation_matcher = re.compile(r"[^a-zA-Z0-9*]")
    # 2. Find all patterns expressions that match regular expression
    matches = regex.finditer(text)
    for m in matches:

        # initialize phi_left and phi_right
        phi_left = False
        phi_right = False

        match_start = m.span()[0]
        match_end = m.span()[1]

        # PHI context left and right
        phi_starts = []
        phi_ends = []
        for start in full_exclude_map:
            phi_starts.append(start)
            phi_ends.append(full_exclude_map[start])

        if match_start in phi_ends:
            phi_left = True

        if match_end in phi_starts:
            phi_right = True

        # Get index of m.group()first alphanumeric character in match
        tokenized_matches = []
        match_text = m.group()
        split_match = re.split(r"(\s+)", re.sub(pre_process, " ", match_text))

        # Get all spans of tokenized match (because remove() function requires tokenized start coordinates)
        coord_tracker = 0
        for element in split_match:
            if element != "":
                if not punctuation_matcher.match(element[0]):
                    current_start = match_start + coord_tracker
                    current_end = current_start + len(element)
                    tokenized_matches.append((current_start, current_end))

                    coord_tracker += len(element)
                else:
                    coord_tracker += len(element)

        # Check for context, and add to coordinate map
        if (
            (context == "left" and phi_left is True)
            or (context == "right" and phi_right)
            or (context == "left_or_right" and (phi_right or phi_left))
            or (context == "left_and_right" and (phi_right and phi_left))
        ):
            for item in tokenized_matches:
                coord_map.add_extend(item[0], item[1])

    return coord_map


def match_all(text, coord_map: CoordinateMap) -> CoordinateMap:
    """ Simply maps to the entirety of the file """
    # add the entire length of the file
    coord_map.add(0, len(text))
    return coord_map


def map_set(text, coord_map: CoordinateMap, pattern: SetFilter) -> CoordinateMap:
    """ Creates a coordinate mapping of words any words in this set"""

    set_data = pattern.data

    # get part of speech we will be sending through this set
    # note, if this is empty we will put all parts of speech through the set
    check_pos = False
    pos_set = set(pattern.pos)
    if len(pos_set) > 0:
        check_pos = True

    cleaned = get_clean(text)
    pos_list = nltk.pos_tag(cleaned)

    start_coordinate = 0
    for tup in pos_list:
        word = tup[0]
        pos = tup[1]
        start = start_coordinate
        stop = start_coordinate + len(word)

        # This converts spaces into empty strings, so we know to skip forward to the next real word
        word_clean = re.sub(r"[^a-zA-Z0-9]+", "", word.lower().strip())
        if len(word_clean) == 0:
            # got a blank space or something without any characters or digits, move forward
            start_coordinate += len(word)
            continue

        if not check_pos or (check_pos and pos in pos_set):
            if word_clean in set_data or word in set_data:
                coord_map.add_extend(start, stop)
            else:
                pass

        # advance our start coordinate
        start_coordinate += len(word)

    return coord_map


def map_pos(text, pattern: PosFilter, coord_map: CoordinateMap) -> CoordinateMap:
    """ Creates a coordinate mapping of words which match this part of speech (POS)"""

    pos_set = set(pattern.pos)

    # Use pre-process to split sentence by spaces AND symbols, while preserving spaces in the split list

    cleaned = get_clean(text)

    pos_list = get_pos(cleaned)  # pos_list = nltk.pos_tag(cleaned)
    start_coordinate = 0
    for tup in pos_list:
        word = tup[0]
        pos = tup[1]
        start = start_coordinate
        stop = start_coordinate + len(word)
        word_clean = re.sub(r"[^a-zA-Z0-9]+", "", word.lower().strip())
        if len(word_clean) == 0:
            # got a blank space or something without any characters or digits, move forward
            start_coordinate += len(word)
            continue

        if pos in pos_set:
            coord_map.add_extend(start, stop)

        # advance our start coordinate
        start_coordinate += len(word)

    return coord_map


def map_ner(
    text,
    pattern: NerFilter,
    coord_map: CoordinateMap,
    stanford_ner_tagger: StanfordNERTagger,
    pre_process=r"[^a-zA-Z0-9]+",
) -> CoordinateMap:
    """ map NER tagging"""
    pos_set = set()
    if pattern.pos:
        pos_set = set(pattern.pos)

    lst = re.split(r"(\s+)", text)
    cleaned = []
    for item in lst:
        if len(item) > 0:
            cleaned.append(item)

    ner_no_spaces = stanford_ner_tagger.tag(cleaned)
    # get our ner tags
    ner_set = {}
    for tup in ner_no_spaces:
        ner_set[tup[0]] = tup[1]
    ner_set_with_locations = {}
    start_coordinate = 0
    for w in cleaned:
        if w in ner_set:
            ner_set_with_locations[w] = (ner_set[w], start_coordinate)
        start_coordinate += len(w)

    # for the text, break into words and mark POS
    # with the parts of speech labeled, match any of these to our coordinate
    # add these coordinates to our coordinate map
    start_coordinate = 0
    for word in cleaned:

        word_clean = re.sub(pre_process, "", word.lower().strip())
        if len(word_clean) == 0:
            # got a blank space or something without any characters or digits, move forward
            start_coordinate += len(word)
            continue

        if word in ner_set_with_locations:
            ner_tag = ner_set_with_locations[word][0]
            start = ner_set_with_locations[word][1]
            if ner_tag in pos_set:
                stop = start + len(word)
                coord_map.add_extend(start, stop)
                print("FOUND: ", word, "NER: ", ner_tag, start, stop)

        # advance our start coordinate
        start_coordinate += len(word)

    return coord_map


def get_exclude_include_maps(
    pattern: Filter,
    txt,
    coord_map: CoordinateMap,
    include_map: CoordinateMap,
    exclude_map: CoordinateMap,
    phi_type_dict: Dict[str, CoordinateMap],
    data_tracker: DataTracker,
):
    exclude = pattern.exclude
    if hasattr(pattern, "filepath"):
        filter_path = pattern.filepath
    else:
        filter_path = pattern.title
    if pattern.phi_type:
        phi_type = pattern.phi_type
    else:
        phi_type = "OTHER"

    for start, stop in coord_map.filecoords():
        if pattern.type != "regex_context":
            if exclude:
                if not include_map.does_overlap(start, stop):
                    exclude_map.add_extend(start, stop)
                    phi_type_dict[phi_type].add_extend(start, stop)

            else:
                if not exclude_map.does_overlap(start, stop):
                    include_map.add_extend(start, stop)
                    data_tracker.non_phi.append(
                        NonPhiEntry(
                            start=start,
                            stop=stop,
                            word=txt[start:stop],
                            filepath=filter_path,
                        )
                    )

        # Add regex_context to map separately
        else:
            if exclude:
                exclude_map.add_extend(start, stop)
                include_map.remove(start, stop)
                phi_type_dict[phi_type].add_extend(start, stop)
            else:
                include_map.add_extend(start, stop)
                exclude_map.remove(start, stop)
                data_tracker.non_phi.append(
                    NonPhiEntry(
                        start=start,
                        stop=stop,
                        word=txt[start:stop],
                        filepath=filter_path,
                    )
                )


def save_to_asterisk(contents, output_file):
    with open(output_file, "w", encoding="utf-8", errors="surrogateescape") as f:
        f.write(contents)


def save_to_i2b2(contents, output_file):
    with open(output_file, "w", errors="xmlcharrefreplace") as f:
        f.write(contents)


def transform_text_asterisk(txt, include_map: CoordinateMap):
    last_marker = 0
    punctuation_matcher = re.compile(r"[^a-zA-Z0-9*]")
    # read the text by character, any non-punc non-overlaps will be replaced
    contents = []
    for i in range(0, len(txt)):

        if i < last_marker:
            continue

        if include_map.does_exist(i):
            # add our preserved text
            start, stop = include_map.get_coords(i)
            contents.append(txt[start:stop])
            last_marker = stop
        elif punctuation_matcher.match(txt[i]):
            contents.append(txt[i])
        else:
            contents.append("*")

    return "".join(contents)


def transform_text_i2b2(tagdata: DataTracker):
    """creates a string in i2b2-XML format"""
    root = "Philter"
    contents = [
        '<?xml version="1.0" ?>\n',
        "<" + root + ">\n",
        "<TEXT><![CDATA[",
        tagdata.text,
        "]]></TEXT>\n",
        "<TAGS>\n",
    ]
    for i, phi in enumerate(tagdata.phi):
        phi_type = phi.phi_type
        contents.append("<")
        contents.append(phi_type)
        contents.append(' id="P')
        contents.append(str(i))
        contents.append('" start="')
        contents.append(str(phi.start))
        contents.append('" end="')
        contents.append(str(phi.stop))
        contents.append('" text="')
        contents.append(phi.word)
        contents.append('" TYPE="')
        contents.append(phi_type)
        contents.append('" comment="" />\n')

    # for loop over complement - PHI, create additional tags (UNKNOWN)
    contents.append("</TAGS>\n")
    contents.append("</" + root + ">\n")

    return "".join(contents)


def detect_encoding(fp):
    if not os.path.exists(fp):
        raise Exception("Filepath does not exist", fp)

    detector = UniversalDetector()
    with open(fp, "rb") as f:
        for line in f:
            detector.feed(line)
            if detector.done:
                break
        detector.close()
    return detector.result


def phi_context(filename, word, word_index, words, context_window=10):
    """ helper function, creates our phi data type with source file, and context window"""
    if not os.path.exists(filename):
        raise Exception("Filepath does not exist", filename)

    left_index = word_index - context_window
    if left_index < 0:
        left_index = 0

    right_index = word_index + context_window
    if right_index >= len(words):
        right_index = len(words) - 1
    window = words[left_index:right_index]

    return {"filename": filename, "phi": word, "context": window}



import argparse
import difflib
import json
import os
import pickle
import re
import sys
import xml.etree.ElementTree as ET
from difflib import SequenceMatcher

import nltk
from chardet.universaldetector import UniversalDetector
from coordinate_map import CoordinateMap


def isolate_phi(xml_folder):
    # isolate all phi and data with coordinates
    # turn them into a json representation
    phi = {}  # fn --> {"text":"...", "phi":[{"type":"DATE"...}]}
    for root_dir, dirs, files in os.walk(xml_folder):
        for f in files:
            with open(root_dir + f, "r", encoding="latin1") as file:
                tree = ET.parse(file)
                root = tree.getroot()

                text = ""
                phi_list = []

                for child in root:
                    if child.tag == "TEXT":
                        text = child.text
                        # if f == '167937985.txt.xml':
                        #     print(text)
                        # print (child.tag, child.attrib, child.text)
                    if child.tag == "TAGS":
                        for t in child:
                            phi_list.append(t.attrib)
                            # print(t.tag, t.attrib, t.text)
                phi[f] = {"text": text, "phi": phi_list}
    return phi


def main():
    # get input/output/filename

    ap = argparse.ArgumentParser()
    ap.add_argument(
        "-x",
        "--xml",
        default="../data/ucsf_xml/",
        help="Path to the directory or the file that contains the note xml files, the default is ../data/ucsf_xml/",
        type=str,
    )
    ap.add_argument(
        "-o",
        "--output",
        default="../data/phi_notes_ucsf.json",
        help="Path to the file that contains a summary of the phi in the xml files, the default is ../data/phi_notes_ucsf.json",
        type=str,
    )
    ap.add_argument(
        "-n",
        "--notes",
        default="../data/ucsf_notes/",
        help="Path to the directory or the file that contains the PHI note, the default is ../data/ucsf_notes/",
        type=str,
    )
    ap.add_argument(
        "-a",
        "--anno",
        default="../data/ucsf_anno/",
        help="Path to the directory or the file that contains the PHI annotation, the default is ../data/ucsf_anno/",
        type=str,
    )
    ap.add_argument(
        "-s",
        "--phi",
        default="../data/ucsf_phi/",
        help="Path to the directory to save the PHI summary in, the default is ../data/ucsf_phi/",
        type=str,
    )
    ap.add_argument(
        "-c",
        "--context",
        default="../data/ucsf_phi_context",
        help="Path to the directory to save the PHI context summary in, the default is ../data/ucsf_phi_context",
        type=str,
    )
    ap.add_argument(
        "-p",
        "--pos",
        default="../data/ucsf_phi_pos",
        help="Path to the directory to save the PHI pos summary in, the default is ../data/ucsf_phi_pos",
        type=str,
    )

    args = ap.parse_args()

    xml_folder = args.xml
    outpath = args.output

    # Run main function
    phi = isolate_phi(xml_folder)

    # save our data
    json.dump(phi, open(outpath, "w"), indent=4)

    NOTES_FOLDER = args.notes
    ANNO_FOLDER = args.anno

    # save our phi notes
    for fn in phi:

        # get text and remove any initial *'s from the raw notes
        txt = phi[fn]["text"].replace("*", " ")

        # save our notes file
        with open(
            NOTES_FOLDER + fn.split(".")[0] + ".txt", "w", encoding="utf-8"
        ) as note_file:
            note_file.write(txt)

        # create a coordinate mapping of all phi
        c = CoordinateMap()
        for p in phi[fn]["phi"]:
            try:
                start = int(p["start"])
                end = int(p["end"])
            except KeyError:
                start = int(p["spans"].split("~")[0])
                end = int(p["spans"].split("~")[1])
            c.add_extend(fn, start, end)

        contents = []
        last_marker = 0
        for start, stop in c.filecoords(fn):
            contents.append(txt[last_marker:start])

            # add a * for each letter preserving shape
            phi_hidden = re.sub(r"[a-zA-Z0-9]", "*", txt[start:stop])
            contents.append(phi_hidden)
            last_marker = stop

        # wrap it up by adding on the remaining values if we haven't hit eof
        if last_marker < len(txt):
            contents.append(txt[last_marker : len(txt)])

        with open(
            ANNO_FOLDER + fn.split(".")[0] + ".txt", "w", encoding="utf-8"
        ) as anno_file:
            anno_file.write("".join(contents))


if __name__ == "__main__":
    main()



import itertools
import re


class CoordinateMap:
    """Hits are stored in a coordinate map data structure

    This class stores start coordinates for any matches found for this pattern"""

    def __init__(self, pattern={"title": "untitled"}, debug=False):
        """ internal data structure maps filepaths to a map of int:string (coordinate start --> stop)

        map is the internal structure of
        { filename : { startcoordinate : stop_coordinate}}
        eg: { "data/foo.txt": {123:126, 19:25} }

        coord2pattern keeps reference of the patterns
        that matched this coorinate (can be multiple patterns)

        all_coords keeps a reference of all coordinates mapped by filename,
        allowing us to easily check if these coordinates have been matched yet"""

        self.map = {}
        self.coord2pattern = {}
        self.pattern = pattern
        self.debug = debug
        self.all_coords = {}

    def add(self, start, stop, overlap=False, pattern=""):
        """ adds a new coordinate to the coordinate map

        if overlap is false, this will reject any overlapping hits (usually from multiple regex scan runs)"""
        if not overlap:
            if self.does_overlap(start, stop):
                return False, "Error, overlaps were found: {} {}".format(start, stop)

        # add our start / stop coordinates
        self.map[start] = stop
        # add these coordinates to our all_coords map
        for i in range(start, stop):
            self.all_coords[i] = 1

        if pattern != "":
            self.add_pattern(start, stop, pattern)
        return True, None

    def add_pattern(self, start, stop, pattern):
        """ adds this pattern to this start coord """
        self.coord2pattern[start] = []
        self.coord2pattern[start].append(pattern)

    def add_extend(self, start, stop, pattern=""):
        """adds a new coordinate to the coordinate map
        if overlaps with another, will extend to the larger size"""
        overlaps = self.max_overlap(start, stop)

        def clear_overlaps(lst):
            for o in lst:
                self.remove(o["orig_start"], o["orig_end"])

        if len(overlaps) == 0:
            # no overlap, just save these coordinates
            self.add(start, stop, pattern=pattern, overlap=True)
            # if filename == "./data/i2b2_notes/167-02.txt":
            # 	print("No overlaps:")
            # 	print(filename,start,stop,pattern)
        elif len(overlaps) == 1:
            clear_overlaps(overlaps)
            # 1 overlap, save this value
            o = overlaps[0]
            self.add(o["new_start"], o["new_stop"], pattern=pattern, overlap=True)
            # if filename == "./data/i2b2_notes/167-02.txt":
            # 	print("One overlap:")
            # 	print(filename,start,stop,pattern)
        else:
            clear_overlaps(overlaps)
            # greater than 1 overlap, by default this is sorted because of scan order
            o1 = overlaps[0]
            o2 = overlaps[-1]
            self.add(o2["new_start"], o1["new_stop"], pattern=pattern, overlap=True)
            # if filename == "./data/i2b2_notes/167-02.txt":
            # 	print("Multiple overlaps:")
            # 	print(filename,start,stop,pattern)

        return True, None

    def remove(self, start, stop):
        """ Removes this coordinate pairing from the map, all_coords, and coord2pattern"""
        # delete from our map structure
        if start in self.map:
            del self.map[start]
        # delete any of these coordinates in our all_coords data structure
        for i in range(start, stop + 1):
            if i in self.all_coords:
                del self.all_coords[i]
        return True, None

    def scan(self):
        """ does an inorder scan of the coordinates and their values"""
        for fn in self.map:
            coords = list(self.map[fn].keys())
            coords.sort()
            for coord in coords:
                yield fn, coord, self.map[fn][coord]

    def keys(self):
        for fn in self.map:
            yield fn

    def get_coords(self, start):
        stop = self.map[start]
        return start, stop

    def filecoords(self):
        """generator does an inorder scan of the coordinates for this file"""
        coords = sorted(self.map.keys())
        for coord in coords:
            yield coord, self.map[coord]

    def does_exist(self, index):
        """ Simple check to see if this index is a hit (start of coordinates)"""
        if index in self.map:
            return True
        return False

    def does_overlap(self, start, stop):
        """ Check if this coordinate overlaps with any existing range"""

        ranges = [list(range(key, self.map[key] + 1)) for key in self.map]
        all_coords = [item for sublist in ranges for item in sublist]
        # removing all_coords implementation until we write some tests
        for i in range(start, stop + 1):
            if i in all_coords:
                return True
        return False

    def calc_overlap(self, start, stop):
        """ given a set of coordinates, will calculate all overlaps
        perf: stop after we know we won't hit any more
        perf: use binary search approach"""

        overlaps = []
        for s in self.map:
            e = self.map[s]
            if s >= start or s <= stop:
                # We found an overlap
                if e <= stop:
                    overlaps.append({"start": s, "stop": e})
                else:
                    overlaps.append({"start": s, "stop": stop})
            elif e >= start or e <= stop:
                if s >= start:
                    overlaps.append({"start": s, "stop": e})
                else:
                    overlaps.append({"start": start, "stop": e})
        return overlaps

    def max_overlap(self, start, stop):
        """ given a set of coordinates, will calculate max of all overlaps
        perf: stop after we know we won't hit any more
        perf: use binary search approach"""

        overlaps = []
        for s in self.map:
            e = self.map[s]
            if s <= start <= e:
                # We found an overlap
                if stop >= e:
                    overlaps.append(
                        {
                            "orig_start": s,
                            "orig_end": e,
                            "new_start": s,
                            "new_stop": stop,
                        }
                    )
                else:
                    overlaps.append(
                        {"orig_start": s, "orig_end": e, "new_start": s, "new_stop": e}
                    )

            elif s <= stop <= e:
                if start <= s:
                    overlaps.append(
                        {
                            "orig_start": s,
                            "orig_end": e,
                            "new_start": start,
                            "new_stop": e,
                        }
                    )
                else:
                    overlaps.append(
                        {"orig_start": s, "orig_end": e, "new_start": s, "new_stop": e}
                    )

        return overlaps

    def get_complement(self, text):
        """ get the complementary coordinates of the input coordinate map (excludes punctuation)"""

        complement_coordinate_map = {}

        current_map_coordinates = []
        for start_key in self.map:
            start = start_key
            stop = self.map[start_key]
            current_map_coordinates += range(start, stop)

        text_coordinates = list(range(0, len(text)))
        complement_coordinates = list(
            set(text_coordinates) - set(current_map_coordinates)
        )

        # Remove punctuation from complement coordinates
        punctuation_matcher = re.compile(r"[^a-zA-Z0-9*]")
        for i in range(0, len(text)):
            if punctuation_matcher.match(text[i]):
                if i in complement_coordinates:
                    complement_coordinates.remove(i)

        # Group complement coordinates into ranges
        def to_ranges(iterable):
            iterable = sorted(set(iterable))
            for key, group in itertools.groupby(
                enumerate(iterable), lambda t: t[1] - t[0]
            ):
                group = list(group)
                yield group[0][1], group[-1][1] + 1

        complement_coordinate_ranges = list(to_ranges(complement_coordinates))

        # Create complement dictionary
        for tup in complement_coordinate_ranges:
            start = tup[0]
            stop = tup[1]
            complement_coordinate_map[start] = stop

        return complement_coordinate_map



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['philter_lite', 'philter_lite.filters.regex', 'philter_lite.generate_dataset']

package_data = \
{'': ['*'],
 'philter_lite': ['configs/*',
                  'data/*',
                  'data/i2b2_anno/*',
                  'data/i2b2_notes/*',
                  'data/i2b2_xml/*',
                  'data/phi/*',
                  'filters/blacklists/*',
                  'filters/regex_context/*',
                  'filters/whitelists/*'],
 'philter_lite.filters.regex': ['addresses/*',
                                'age/*',
                                'contact/*',
                                'dates/*',
                                'mrn_id/*',
                                'organizations/*',
                                'safe/*',
                                'salutations/*',
                                'ucsf_regex/*']}

install_requires = \
['chardet>=3.0.4,<4.0.0', 'nltk>=3.5,<4.0', 'xmltodict>=0.12.0,<0.13.0']

entry_points = \
{'console_scripts': ['philter_lite = philter_lite.main:main']}

setup_kwargs = {
    'name': 'philter-lite',
    'version': '0.1.1',
    'description': '',
    'long_description': None,
    'author': 'Beau Norgeot',
    'author_email': 'beaunorgeot@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import argparse
import distutils.util
import os
import re

from philter_lite import evaluate, philter


def main():
    # get input/output/filename
    help_str = """ Philter -- PHI filter for clinical notes """
    ap = argparse.ArgumentParser(description=help_str)
    ap.add_argument(
        "-i",
        "--input",
        default="./data/i2b2_notes/",
        help="Path to the directory or the file that contains the PHI note, the default is "
        "./data/i2b2_notes/",
        type=str,
    )
    ap.add_argument(
        "-a",
        "--anno",
        default="./data/i2b2_anno/",
        help="Path to the directory or the file that contains the PHI annotation, the default is "
        "./data/i2b2_anno/",
        type=str,
    )
    ap.add_argument(
        "-o",
        "--output",
        default="./data/i2b2_results/",
        help="Path to the directory to save the PHI-reduced notes in, the default is ./data/i2b2_results/",
        type=str,
    )
    ap.add_argument(
        "-f",
        "--filters",
        default="./configs/integration_1.json",
        help="Path to our config file, the default is ./configs/integration_1.json",
        type=str,
    )
    ap.add_argument(
        "-x",
        "--xml",
        default="./data/phi_notes.json",
        help="Path to the json file that contains all xml data",
        type=str,
    )
    ap.add_argument(
        "-c",
        "--coords",
        default="./data/coordinates.json",
        help="Path to the json file that contains the coordinate map data",
        type=str,
    )
    ap.add_argument(
        "--eval_output",
        default="./data/phi/",
        help="Path to the directory that the detailed eval files will be outputted to",
        type=str,
    )
    ap.add_argument(
        "-v",
        "--verbose",
        default=True,
        help="When verbose is true, will emit messages about script progress",
        type=lambda x: bool(distutils.util.strtobool(x)),
    )
    ap.add_argument(
        "-e",
        "--run_eval",
        default=True,
        help="When run_eval is true, will run our eval script and emit summarized results to terminal",
        type=lambda x: bool(distutils.util.strtobool(x)),
    )
    ap.add_argument(
        "-t",
        "--freq_table",
        default=False,
        help="When freqtable is true, will output a unigram/bigram frequency table of all note words and "
        "their PHI/non-PHI counts",
        type=lambda x: bool(distutils.util.strtobool(x)),
    )
    ap.add_argument(
        "-n",
        "--initials",
        default=True,
        help="When initials is true, will include initials PHI in recall/precision calculations",
        type=lambda x: bool(distutils.util.strtobool(x)),
    )
    ap.add_argument(
        "--outputformat",
        default="i2b2",
        help='Define format of annotation, allowed values are "asterisk", "i2b2". Default '
        'is "asterisk"',
        type=str,
    )
    ap.add_argument(
        "--ucsfformat",
        default=False,
        help="When ucsfformat is true, will adjust eval script for slightly different xml format",
        type=lambda x: bool(distutils.util.strtobool(x)),
    )
    ap.add_argument(
        "--prod",
        default=False,
        help="When prod is true, this will run the script with output in i2b2 xml format without running "
        "the eval script",
        type=lambda x: bool(distutils.util.strtobool(x)),
    )
    ap.add_argument(
        "--cachepos",
        default=None,
        help="Path to a directoy to store/load the pos data for all notes. If no path is specified then "
        "memory caching will be used.",
        type=str,
    )

    args = ap.parse_args()
    run_eval = args.run_eval
    verbose = args.verbose

    if args.prod:
        run_eval = False
        verbose = False

        philter_config = {
            "verbose": verbose,
            "run_eval": run_eval,
            "finpath": args.input,
            "foutpath": args.output,
            "outformat": args.outputformat,
            "filters": args.filters,
            "cachepos": args.cachepos,
        }

    else:
        philter_config = {
            "verbose": args.verbose,
            "run_eval": args.run_eval,
            "freq_table": args.freq_table,
            "initials": args.initials,
            "finpath": args.input,
            "foutpath": args.output,
            "outformat": args.outputformat,
            "ucsfformat": args.ucsfformat,
            "anno_folder": args.anno,
            "filters": args.filters,
            "xml": args.xml,
            "coords": args.coords,
            "eval_out": args.eval_output,
            "cachepos": args.cachepos,
        }

    if verbose:
        print("RUNNING ", philter_config["filters"])

    filters = philter.build_filters(philter_config["filters"])

    stanford_ner = None
    if len([x for x in filters if x.type == "stanford_ner"]):
        stanford_ner = philter.build_ner_tagger(
            philter_config["stanford_ner_tagger"]["classifier"],
            philter_config["stanford_ner_tagger"]["stanford_ner_tagger"],
        )

    for root, dirs, files in os.walk(philter_config["finpath"]):
        for file in files:
            with open(os.path.join(root, file)) as inf:
                entry, include_map, exclude_map, data_tracker = philter.map_coordinates(
                    inf.read(), patterns=filters
                )
                if philter_config["outformat"] == "i2b2":
                    with open(
                        os.path.join(philter_config["foutpath"], f"{file}.txt"), "w"
                    ) as fout:
                        fout.write(philter.transform_text_i2b2(data_tracker))
                elif philter_config["outformat"] == "asterisk":
                    with open(
                        os.path.join(philter_config["foutpath"], f"{file}.txt"), "w"
                    ) as fout:
                        fout.write(philter.transform_text_asterisk(entry, include_map))

    # evaluate the effectiveness
    if run_eval and args.outputformat == "asterisk":
        evaluate.eval(
            in_path=args.output,
            anno_path=args.anno,
            anno_suffix=".txt",
            fn_output="data/phi/fn.txt",
            fp_output="data/phi/fp.txt",
            summary_output="./data/phi/summary.json",
            punctuation_matcher=re.compile(r"[^a-zA-Z0-9\*]"),
        )


if __name__ == "__main__":
    main()
